import torch
import torchvision
import numpy as np
from numpy.random import randint
from typing import Tuple, Optional, Dict, List
from torch.utils.data import DataLoader, Dataset, Subset, ConcatDataset
from PIL import Image


class Benchmark:
    """
    Base class for continual learning benchmarks.
    It implements logic for loading/serving continual learning datasets.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):
        """
        Args:
            num_tasks: The number of tasks for the benchmark.
            per_task_examples: If set, each task will include part of the original benchmark rather than full data.
            per_task_joint_examples: If set, the benchmark will support joint/multitask loading of tasks.
            per_task_memory_examples: If set, the benchmark will support episodic memory/replay buffer loading of tasks.
            per_task_subset_examples: If set, the benchmark will support loading a pre-defined subset of each task.
            task_input_transforms: If set, the benchmark will use the provided torchvision transform.
            task_target_transforms: If set, the benchmark will use the provided target transform for targets.
            
        . note::
            If :attr:`task_input_transforms` or :attr:`task_target_transforms`, they should be a list
            of size `num_tasks` where each element of the list can be a torchvision (Composed) transform.
        """
        
        # Task details
        self.num_tasks = num_tasks
        self.per_task_seq_examples = per_task_examples
        self.per_task_joint_examples = per_task_joint_examples
        self.per_task_memory_examples = per_task_memory_examples
        self.per_task_subset_examples = per_task_subset_examples
        
        # Optional transformations
        self.task_input_transforms = task_input_transforms
        self.task_target_transforms = task_target_transforms

        # Book-keeping variables: mostly used for storing indices of data points for each task
        self.trains, self.tests = {}, {}
        self.joint_indices_train, self.joint_indices_test = {}, {}
        self.memory_indices_train, self.memory_indices_test = {}, {}
        self.seq_indices_train, self.seq_indices_test = {}, {}
        self.subset_indices_train, self.subset_indices_test = {}, {}
        self.sanity_check_inputs()
    
    def sanity_check_inputs(self):
        """
        A method for sanity checking arguments.
        E.g., checking if all provided arguments are valid.
        """
        self.sanity_check_transforms()
    
    def sanity_check_transforms(self):
        if self.task_input_transforms is not None and len(self.task_input_transforms) != self.num_tasks:
            raise ValueError("task_input_transform is either None (using default transform),\
                             or should be a list of size `num_tasks` that provides transforms for each task")
            
    def prepare_datasets(self):
        """
        Prepares datasets: will be called after `load_datasets`.
        Responsible for computing index for various methods. E.g., selecting subset/memory indices for each task.
        """
        if self.per_task_joint_examples:
            self.precompute_joint_indices()
        if self.per_task_memory_examples:
            self.precompute_memory_indices()
        if self.per_task_seq_examples:
            self.precompute_seq_indices()
        if self.per_task_subset_examples:
            self.precompute_subset_indices()
   
    def load_datasets(self):
        """
        Loading datasets from file.
        """
        raise NotImplementedError
    
    def precompute_joint_indices(self):
        """
        For each task, (randomly) computes the indices of the subset of data points in the task's dataset.
        Then, once `load_joint()` method is called, uses these indices to return a PyTorch `Subset` dataset.
        .. note:: This method will be called only if the benchmark is initiated with `per_task_joint_examples`.
        """
        for task in range(1, self.num_tasks+1):
            self.joint_indices_train[task] = randint(0, len(self.trains[task]), size=self.per_task_joint_examples)
            self.joint_indices_test[task] = randint(0, len(self.tests[task]), size=self.per_task_joint_examples)
            
    def precompute_memory_indices(self):
        """
        For each task, (randomly) computes the indices of the subset of data points in the task's dataset.
        Then, once `load_memory()` method is called, uses these indices to return a PyTorch `Subset` dataset.
        .. note:: This method will be called only if the benchmark is initiated with `per_task_memory_examples`.
        """
        raise NotImplementedError
    
    def precompute_subset_indices(self):
        for task in range(1, self.num_tasks + 1):
            self.subset_indices_train[task] = randint(0, len(self.trains[task]), size=self.per_task_subset_examples)
            self.subset_indices_test[task] = randint(0, len(self.tests[task]), size=min(self.per_task_subset_examples, len(self.tests[task])))

    def precompute_seq_indices(self):
        if self.per_task_seq_examples > len(self.trains[1]):
            raise ValueError(f"per task examples = {self.per_task_seq_examples} but first task's examples = {len(self.trains[1])}")
        
        for task in range(1, self.num_tasks+1):
            self.seq_indices_train[task] = randint(0, len(self.trains[task]), size=self.per_task_seq_examples)
            self.seq_indices_test[task] = randint(0, len(self.tests[task]), size=min(self.per_task_seq_examples, len(self.tests[task])))
    
    def _calculate_num_examples_per_class(self, start_class, end_class, num_samples):
        num_classes = end_class - start_class + 1
        num_examples_per_class = num_samples//num_classes
        result = [num_examples_per_class]*num_classes
        
        # if memory_size can't be divided by num_class classes
        # e.g., memory_size is 32, but we have 5 classes.
        if num_classes * num_examples_per_class < num_samples:
            # how many more examples we need?
            diff = num_samples - (num_classes * num_examples_per_class)
            # add examples
            while diff:
                diff -= 1
                result[randint(0, num_classes)] += 1
        return result
   
    def sample_uniform_class_indices(self, dataset, start_class, end_class, num_samples) -> List:
        """
        Selects a subset of size `num_samples` from `start_class` to `end_class`.
        Args:
            dataset: The input dataset that the samples will be drawn from.
            start_class: The start_class (inclusive)
            end_class: The end_classes (inclusive)
            num_samples: Number of samples.

        Returns:
            Indices of the selected samples.
    
        .. note:: This method is specially useful for split datasets. E.g., selecting classes 0 to 4 for Split-CIFAR-100.
        
        .. warning:: If `num_samples > len(dataset)`, then the output`s shape will be equal to `len(dataset)`.
        """
        target_classes = dataset.targets.clone().detach().numpy()
        num_examples_per_class = self._calculate_num_examples_per_class(start_class, end_class, num_samples)
        class_indices = []
        # choose num_examples_per_class for each class
        for i, cls_number in enumerate(range(start_class, end_class+1)):
            target = (target_classes == cls_number)
            #  maybe that class doesn't exist
            num_candidate_examples = len(np.where(target == 1)[0])
            if num_candidate_examples:
                selected_indices = np.random.choice(np.where(target == 1)[0],
                                                    min(num_candidate_examples, num_examples_per_class[i]),
                                                    replace=False)
                class_indices += list(selected_indices)
        return class_indices

    def load(self,
             task: int,
             batch_size: int,
             shuffle: Optional[bool] = True,
             num_workers: Optional[int] = 0,
             pin_memory: Optional[bool] = True) -> Tuple[DataLoader, DataLoader]:
        
        """
        Makes train/val dataloaders for a specific task.
        Args:
            task: The task number.
            batch_size: The batch_size for dataloaders.
            shuffle: Should loaders be shuffled? Default: True.
            num_workers: corresponds to Pytorch's `num_workers` argument. Default: 0
            pin_memory: corresponds to Pytorch's `pin_memory` argument. Default: True.

        Returns:
            a Tuple of dataloaders, i.e., (train_loader, validation_loader).
            
        Examples::
            >>> benchmark = Benchmark(num_tasks=2)
            >>> # task 1 loaders
            >>> train_loader_1, val_loader_1 = benchmark.load(1, batch_size=32)
            >>> # task 2 loaders
            >>> train_loader_2, val_loader_2 = benchmark.load(2, batch_size=64)
        """
        
        if task > self.num_tasks:
            raise ValueError(f"Asked to load task {task} but the benchmark has {self.num_tasks} tasks")
        
        if self.per_task_seq_examples:
            trainset = Subset(self.trains[task], self.seq_indices_train[task])
            testset = Subset(self.tests[task], self.seq_indices_test[task])
        else:
            trainset = self.trains[task]
            testset = self.tests[task]
        train_loader = DataLoader(trainset, batch_size, shuffle, num_workers=num_workers,
                                  pin_memory=pin_memory)
        test_loader = DataLoader(testset, 256, True, num_workers=num_workers, pin_memory=pin_memory)
    
        return train_loader, test_loader

    def load_joint(self,
                   task: int,
                   batch_size: int,
                   shuffle: Optional[bool] = True,
                   num_workers: Optional[int] = 0,
                   pin_memory: Optional[bool] = True) -> Tuple[DataLoader, DataLoader]:
        """
        Makes dataloaders for joint/multitask settings.
        i.e., for task `t` returns datasets for tasks `1, 2, ..., t-1, t`.
        
        Args:
            task: The task number.
            batch_size: The batch_size for dataloaders.
            shuffle: Should loaders be shuffled? Default: True.
            num_workers: corresponds to Pytorch's `num_workers` argument. Default: 0
            pin_memory: corresponds to Pytorch's `pin_memory` argument. Default: True.

        Returns:
            a Tuple of dataloaders, i.e., (train_loader, validation_loader).
            
        Examples::
            >>> benchmark = Benchmark(num_tasks=2, per_task_joint_examples=128)
            >>> # task 1 loaders (single): returns 4 batches (i.e., 128 examples)
            >>> train_loader_1, val_loader_1 = benchmark.load(1, batch_size=32)
            >>> # task 1 loaders (joint): returns 4 batches (i.e., 128 examples)
            >>> joint_train_loader_1, joint_val_loader_1 = benchmark.load_joint(1, batch_size=32)
            >>> # task 1 loaders (single): returns 4 batches (i.e., 128 examples)
            >>> train_loader_2, val_loader_2 = benchmark.load(2, batch_size=32)
            >>> # task 1 loaders (single): returns 8 batches (i.e., 256 examples)
            >>> joint_train_loader_2, joint_val_loader_2 = benchmark.load(2, batch_size=32)
        
        .. warning::
            The method will throw an error if `Benchmark` is instantiated without `per_task_joint_examples`.
            The reason is that, behind the scenese, we compute the indices for joint examples in
            `precompute_joint_indices()` method and this method relies on that computations.
        """
        if not self.per_task_joint_examples:
            raise ValueError("Called load_joint() but per_task_joint_examples is not set")
        
        if task > self.num_tasks:
            raise ValueError(f"Asked to load task {task} but the benchmark has {self.num_tasks} tasks")
    
        trains, tests = [], []
        for prev_task in range(1, task + 1):
            prev_train = Subset(self.trains[prev_task], self.joint_indices_train[prev_task])
            prev_test = Subset(self.tests[prev_task], self.joint_indices_test[prev_task])
            trains.append(prev_train)
            tests.append(prev_test)
    
        trains, tests = ConcatDataset(trains), ConcatDataset(tests)
        train_loader = DataLoader(trains, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        test_loader = DataLoader(tests, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        return train_loader, test_loader

    def load_subset(self,
                    task: int,
                    batch_size: int,
                    shuffle: Optional[bool] = True,
                    num_workers: Optional[int] = 0,
                    pin_memory: Optional[bool] = True) -> Tuple[DataLoader, DataLoader]:
        if not self.per_task_subset_examples:
            raise ValueError("Called load_subset() without setting per_task_subset_examples")
        subset_train = Subset(self.trains[task], self.subset_indices_train[task])
        subset_test = Subset(self.tests[task], self.subset_indices_test[task])
        train_loader = DataLoader(subset_train, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        test_loader = DataLoader(subset_test, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        return train_loader, test_loader

    def load_memory(self,
                    task: int,
                    batch_size: int,
                    shuffle: Optional[bool] = True,
                    num_workers: Optional[int] = 0,
                    pin_memory: Optional[bool] = True) -> Tuple[DataLoader, DataLoader]:
        """
        Makes dataloaders for episodic memory/replay buffer.
        
        Args:
            task: The task number.
            batch_size: The batch_size for dataloaders.
            shuffle: Should loaders be shuffled? Default: True.
            num_workers: corresponds to Pytorch's `num_workers` argument. Default: 0
            pin_memory: corresponds to Pytorch's `pin_memory` argument. Default: True.

        Returns:
            a Tuple of dataloaders, i.e., (train_loader, validation_loader).

        Examples::
            >>> benchmark = Benchmark(num_tasks=2, per_task_memory_examples=16)
            >>> # task 1 memory loaders: returns 2 batches (i.e., 16 examples)
            >>> mem_train_loader_1, mem_val_loader_1 = benchmark.load_memory(1, batch_size=8)
            >>> # task 2 memory loaders: returns 4 batches (i.e., 16 examples)
            >>> mem_train_loader_2, mem_val_loader_2 = benchmark.load_memory(2, batch_size=4)

        .. note::
            This method uses `class_uniform` sampling.  i.e., if each task has 10 classes,
            and `per_task_memory_examples=20`, then the returend samples have 2 examples per class.
            
        .. warning::
            The method will throw an error if `Benchmark` is instantiated without :attr:`per_task_memory_examples`.
            The reason is that, behind the scenese, we compute the indices for memory examples in
            `precompute_memory_indices()` method and this method relies on that computations.
        """
        if not self.per_task_memory_examples:
            raise ValueError("Called load_memory() but per_task_memory_examples is not set")
        
        if task > self.num_tasks:
            raise ValueError(f"Asked for memory of task={task} while the benchmark has {self.num_tasks} tasks")
    
        train_indices = self.memory_indices_train[task]
        test_indices = self.memory_indices_test[task]
        train_dataset = Subset(self.trains[task], train_indices)
        test_dataset = Subset(self.tests[task], test_indices)
    
        train_loader = DataLoader(train_dataset, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        test_loader = DataLoader(test_dataset, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
    
        return train_loader, test_loader

    def load_memory_joint(self,
                          task: int,
                          batch_size: int,
                          shuffle: Optional[bool] = True,
                          num_workers: Optional[int] = 0,
                          pin_memory: Optional[bool] = True) -> Tuple[DataLoader, DataLoader]:
        if task > self.num_tasks:
            raise ValueError(f"Asked to load memory of task={task} but the benchmark has {self.num_tasks} tasks")
        trains, tests = [], []
        for t in range(1, task + 1):
            train_indices = self.memory_indices_train[t]
            test_indices = self.memory_indices_test[t]
            train_dataset = Subset(self.trains[t], train_indices)
            test_dataset = Subset(self.tests[t], test_indices)
            trains.append(train_dataset)
            tests.append(test_dataset)
    
        trains, tests = ConcatDataset(trains), ConcatDataset(tests)
        train_loader = DataLoader(trains, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        test_loader = DataLoader(tests, batch_size, shuffle, num_workers=num_workers, pin_memory=pin_memory)
        return train_loader, test_loader


class DynamicTransformDataset(Dataset):
    """
    A lightweight wrapper around PyTorch dataset.
    Simply, applies custom transforms on the dataset, rather than loading datasets again.
    """
    def __init__(self, task_id: int, dataset: Dataset, input_transform=None, target_transform=None):
        self.task_id = task_id
        self.dataset = dataset
        self.input_transform = input_transform
        self.target_transform = target_transform
    
    def __getitem__(self, index):
        img, target = self.dataset.data[index], int(self.dataset.targets[index])
        img = Image.fromarray(img.numpy(), mode='L')
        
        # transforms
        if self.input_transform is not None:
            img = self.input_transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        
        return img, target, self.task_id
    
    def __len__(self) -> int:
        return len(self.dataset.data)


class SplitDataset(Dataset):
    """
    A lightweight wrapper around PyTorch dataset for split benchmarks.
    """
    def __init__(self, task_id, classes_per_split, dataset):
        self.inputs = []
        self.targets = []
        self.task_id = task_id
        self.classes_per_split = classes_per_split
        self.__build_split(dataset, task_id)

    def __build_split(self, dataset, task_id):
        start_class = (task_id-1) * self.classes_per_split
        end_class = task_id * self.classes_per_split
        # For CIFAR-like datasets in torchvision where targets are list
        if isinstance(dataset.targets, list):
            target_classes = np.asarray(dataset.targets)
        # for MNIST-like datasets where targets are tensors
        else:
            target_classes = dataset.targets.clone().detach().numpy()
        # target_classes = dataset.targets.clone().detach().numpy()
        selected_indices = np.where(np.logical_and(start_class <= target_classes, target_classes < end_class))[0]
        for idx in selected_indices:
            img, target = dataset[idx]
            target = torch.tensor(target)
            self.inputs.append(img)
            self.targets.append(target)
        
        self.inputs = torch.stack(self.inputs)
        self.targets = torch.stack(self.targets)

    def __getitem__(self, index: int):
        img, target = self.inputs[index], int(self.targets[index])
        return img, target, self.task_id

    def __len__(self) -> int:
        return len(self.inputs)


import os
import torch
import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from cl_gym.benchmarks import Benchmark
from typing import Optional, Tuple
from torchvision.datasets.utils import download_and_extract_archive

PAMAP_ZIP_URL = "https://www.imirzadeh.me/dl/PAMAP2.zip"
DOWNLOAD_DIR = './data/'


class PAMAP2Loader:
    def __init__(self, subject_id, window_seconds=2.56, overlap=0.5, location=None, test_ratio=0.25):
        download_and_extract_archive(PAMAP_ZIP_URL, DOWNLOAD_DIR)
        self.subject_id = subject_id
        self.frequency_Hz = 100
        self.window_seconds = window_seconds
        self.overlap = overlap
        self.location = location.lower() if location else location
        self.test_ratio = test_ratio
        self.raw_data = []
        self.check_inputs()
        self.load_data()
        self.data = self.extract_windows()
    
    def check_inputs(self):
        if not 1 <= self.subject_id <= 8:
            raise ValueError("PAMAP2 has 8 subjects (1 to 8)")
        if not 0 <= self.overlap < 1:
            raise ValueError("Overlap is between 0 and 1 (i.e., relative to window size)")
        if self.location and self.location not in ['chest', 'hand', 'ankle']:
            raise ValueError("PAMAP2 supposrts 'chest', 'hand', and 'ankle' for locations")
    
    def _get_columns(self):
        columns = ['time', 'activity',
                   'hand:accel_X', 'hand:accel_Y', 'hand:accel_Z',
                   'hand:gyro_X', 'hand:gyro_Y', 'hand:gyro_Z',
                   'chest:accel_X', 'chest:accel_Y', 'chest:accel_Z',
                   'chest:gyro_X', 'chest:gyro_Y', 'chest:gyro_Z',
                   'ankle:accel_X', 'ankle:accel_Y', 'ankle:accel_Z',
                   'ankle:gyro_X', 'ankle:gyro_Y', 'ankle:gyro_Z']
        if self.location:
            columns = ['time', 'activity']
            for axis in ["accel_X", 'accel_Y', 'accel_Z', 'gyro_X', 'gyro_Y', 'gyro_Z']:
                columns.append(f"{self.location}:{axis}")
        return columns
        
    def load_data(self):
        filepath = os.path.join(DOWNLOAD_DIR, 'PAMAP2', '{}.csv'.format(self.subject_id))
        self.raw_data = pd.read_csv(filepath)
    
    def _calculate_window_range(self, idx):
        window_steps = int(self.window_seconds * self.frequency_Hz)
        overlap_steps = int(window_steps * self.overlap)
        window_start = max(idx * window_steps - overlap_steps, 0)
        window_end = window_start + window_steps
        return window_start, window_end
    
    @staticmethod
    def correct_labels(label):
        # PAMAP labels: [ 1  2  3 17 16 12 13  4  6  7  5 24]
        if 1 <= label <= 7:
            return label - 1
        else:
            map = {12: 7, 13: 8, 16: 9, 17: 10, 24: 11}
            return map[label]
    
    def extract_windows(self):
        columns = self._get_columns()
        df = self.raw_data[columns].sort_values(by="time").drop("time", axis=1)
        df[['activity']] = df[['activity']].applymap(lambda x: self.correct_labels(x))
        idx = 0
        windows = []
        while True:
            start, end = self._calculate_window_range(idx)
            if end > len(df)-1:
                break
            window = df.iloc[start:end]
            windows.append(window)
            idx += 1
        return windows
    
    def create_train_test(self):
        data = self.data
        # shuffle
        np.random.shuffle(data)
        
        # separate train and test
        num_trains = int((1-self.test_ratio) * len(data))
        trains = data[:num_trains]
        tests = data[num_trains:]
        return trains, tests
        

class PAMAP2Dataset(Dataset):
    def __init__(self, data):
        # data is list of pandas df
        inputs = []
        targets = []
        for d in data:
            inputs.append(d.drop('activity', axis=1).fillna(method='ffill', axis=1).fillna(value=0).to_numpy(dtype=np.float32)/10.0)
            targets.append(d.iloc[0].activity)
        
        inputs = np.array(inputs)
        targets = np.array(targets, dtype=np.long)
        self.inp = torch.from_numpy(inputs)
        self.targets = torch.from_numpy(targets)
    
    def __len__(self):
        return len(self.inp)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        return self.inp[idx, :], self.targets[idx], [None]*len(idx)


class PAMAP2(Benchmark):
    """
    PAMAP2 benchmark: activity recognition with at most 8 tasks.
    This is a time-series benchmark.
    """
    def __init__(self,
                 num_tasks: int = 8,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 shuffle_subjects=True):
        # TODO: Add window-size and overlap to benchmark
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples)
        self.shuffle_subjects = shuffle_subjects
        self.load_datasets()
        self.prepare_datasets()
    
    def load_datasets(self):
        subject_order = list(range(1, 9))
        if self.shuffle_subjects:
            np.random.shuffle(subject_order)
            
        for task in range(1, self.num_tasks+1):
            trains, tests = PAMAP2Loader(subject_order[task-1]).create_train_test()
            self.trains[task] = PAMAP2Dataset(trains)
            self.tests[task] = PAMAP2Dataset(tests)
            
    def precompute_memory_indices(self):
        for task in range(1, self.num_tasks + 1):
            indices_train = self.sample_uniform_class_indices(self.trains[task], 0, 11, self.per_task_memory_examples)
            indices_test = self.sample_uniform_class_indices(self.tests[task], 0, 11, self.per_task_memory_examples)
            # assert len(indices_train) == len(indices_test)
            self.memory_indices_train[task] = indices_train[:]
            self.memory_indices_test[task] = indices_test[:]


from cl_gym.utils.callbacks.base import ContinualCallback
from cl_gym.utils.callbacks.model import BackboneFreeze
from cl_gym.utils.callbacks.model import ModelCheckpoint
from cl_gym.utils.callbacks.metric_manager import MetricCollector
from cl_gym.utils.callbacks.experiment_manager import ExperimentManager

__all__ = ["ContinualCallback",
           "BackboneFreeze",
           "ModelCheckpoint",
           "MetricCollector",
           "ExperimentManager",
           ]


import torch
import numpy as np
import torch.nn as nn


def flatten_weights(model, numpy_output: bool = True):
    """
    Flattens a PyTorch model. i.e., concat all parameters as a single, large vector.
    :param model: PyTorch model
    :param numpy_output: should the output vector be casted as numpy array?
    :return: the flattened (vectorized) model parameters either as Numpy array or Torch tensors
    """
    all_params = []
    for param in model.parameters():
        all_params.append(param.view(-1))
    all_params = torch.cat(all_params)
    if numpy_output:
        return all_params.cpu().detach().numpy()
    return all_params


def flatten_grads(model):
    """
    Flattens the gradients of a model (after `.backward()` call) as a single, large vector.
    :param model: PyTorch model.
    :return: 1D torch Tensor
    """
    all_grads = []
    for name, param in model.named_parameters():
        all_grads.append(param.grad.view(-1))
    return torch.cat(all_grads)


def assign_weights(model, weights):
    """
    Manually assigns `weights` of a Pytorch `model`.
    Note that weights is of vector form (i.e., 1D array or tensor).
    Usage: For implementation of Mode Connectivity SGD algorithm.
    :param model: Pytorch model.
    :param weights: A flattened (i.e., 1D) weight vector.
    :return: The `model` updated with `weights`.
    """
    state_dict = model.state_dict(keep_vars=True)
    # The index keeps track of location of current weights that is being un-flattened.
    index = 0
    # just for safety, no grads should be transferred.
    with torch.no_grad():
        for param in state_dict.keys():
            # ignore batchnorm params
            if 'running_mean' in param or 'running_var' in param or 'num_batches_tracked' in param:
                continue
            param_count = state_dict[param].numel()
            param_shape = state_dict[param].shape
            state_dict[param] = nn.Parameter(torch.from_numpy(weights[index:index+param_count].reshape(param_shape)))
            index += param_count
    model.load_state_dict(state_dict)
    return model


def assign_grads(model, grads):
    """
    Similar to `assign_weights` but this time, manually assign `grads` vector to a model.
    :param model: PyTorch Model.
    :param grads: Gradient vectors.
    :return:
    """
    state_dict = model.state_dict(keep_vars=True)
    index = 0
    for param in state_dict.keys():
        # ignore batchnorm params
        if 'running_mean' in param or 'running_var' in param or 'num_batches_tracked' in param:
            continue
        param_count = state_dict[param].numel()
        param_shape = state_dict[param].shape
        state_dict[param].grad =  grads[index:index+param_count].view(param_shape).clone()
        index += param_count
    model.load_state_dict(state_dict)
    return model


import os
import numpy as np
from typing import Optional, Callable, Dict
from cl_gym.utils.loggers import Logger
from cl_gym.utils.callbacks import ContinualCallback
from cl_gym.utils.callbacks.helpers import IntervalCalculator, Visualizer
from cl_gym.utils.metrics import ContinualMetric, PerformanceMetric, ForgettingMetric


class MetricCollector(ContinualCallback):
    """
    Collects metrics during the learning.
    This callback can support various metrics such as average accuracy/error, and average forgetting.
    """
    def __init__(self, num_tasks: int,
                 epochs_per_task: Optional[int] = 1,
                 collect_on_init: bool = False,
                 collect_metrics_for_future_tasks: bool = False,
                 eval_interval: str = 'epoch',
                 eval_type: str = 'classification',
                 tuner_callback: Optional[Callable[[float, bool], None]] = None):
        """
        
        Args:
            num_tasks: The number of task for the learning experience.
            epochs_per_task: The number of epochs per task.
            collect_on_init: Should also collect metrics before training starts?
            collect_metrics_for_future_tasks: Should collect metrics for future tasks? (e.g.,, for forward-transfer)
            eval_interval: The intervals at which the algorithm will be evaluated. Can be either `task` or `epoch`
            eval_type: Is this a `classification` task or `regression` task?
            tuner_callback: Optional tuner callback than can be called with eval metrics for parameter optimization.
        """
        self.num_tasks = num_tasks
        self.epochs_per_task = epochs_per_task
        self.collect_on_init = collect_on_init
        self.collect_metrics_for_future_tasks = collect_metrics_for_future_tasks
        self.eval_interval = eval_interval.lower()
        self.tuner_callback = tuner_callback
        self.save_dirs = ['plots', 'metrics']
        self.eval_type = eval_type.lower()
        self.meters = self._prepare_meters()
        self.interval_calculator = IntervalCalculator(self.num_tasks, self.epochs_per_task, self.eval_interval)
        super(MetricCollector, self).__init__("MetricCollector", self.save_dirs)
        self._verify_inputs()

    def _prepare_meters(self) -> Dict[str, ContinualMetric]:
        if self.eval_type == 'classification':
            return {'accuracy': PerformanceMetric(self.num_tasks, self.epochs_per_task),
                    'forgetting': ForgettingMetric(self.num_tasks, self.epochs_per_task),
                    'loss': PerformanceMetric(self.num_tasks, self.epochs_per_task)}
        else:
            return {'loss': PerformanceMetric(self.num_tasks, self.epochs_per_task)}
        
    def _verify_inputs(self):
        if self.num_tasks <= 0 or self.epochs_per_task <= 0:
            raise ValueError("'num_tasks' and 'epochs_per_task' should be greater than 0")
        if self.eval_interval not in ['epoch', 'task']:
            raise ValueError("'eval_interval' should be either 'task' or 'epoch'")
        if self.eval_type not in ['classification', 'regression']:
            raise ValueError("'eval_type' for metrics should be either 'classification' or 'regression'")

    def _update_meters(self, task_learned: int, task_evaluated: int, metrics: dict, relative_step: int):
        if self.eval_type == 'classification':
            self.meters['loss'].update(task_learned, task_evaluated, metrics['loss'], relative_step)
            self.meters['accuracy'].update(task_learned, task_evaluated, metrics['accuracy'], relative_step)
            self.meters['forgetting'].update(task_learned, task_evaluated, metrics['accuracy'], relative_step)
        else:
            self.meters['loss'].update(task_learned, task_evaluated, metrics['loss'], relative_step)
    
    def _update_logger(self, trainer, task_evaluated: int, metrics: dict, global_step: int):
        if trainer.logger is None:
            return
        
        if self.eval_type == 'classification':
            trainer.logger.log_metric(f'acc_{task_evaluated}', round(metrics['accuracy'], 2), global_step)
            trainer.logger.log_metric(f'loss_{task_evaluated}', round(metrics['loss'], 2), global_step)
            if trainer.current_task > 0:
                avg_acc = round(self.meters['accuracy'].compute(trainer.current_task), 2)
                trainer.logger.log_metric(f'average_acc', avg_acc, global_step)
        else:
            trainer.logger.log_metric(f'loss_{task_evaluated}', round(metrics['loss'], 2), global_step)
            if trainer.current_task > 0:
                avg_loss = round(self.meters['loss'].compute(trainer.current_task), 5)
                trainer.logger.log_metric(f'average_loss', avg_loss, global_step)

    def _update_tuner(self, is_final_score: bool):
        if self.tuner_callback is None:
            return
        if self.eval_type == 'classification':
            score = self.meters['accuracy'].compute_final()
        else:
            score = self.meters['loss'].compute_final()
        self.tuner_callback(score, is_final_score)
    
    def log_metrics(self, trainer, task_learned: int, task_evaluated: int,
                     metrics: dict, global_step: int, relative_step: int):
        self._update_meters(task_learned, task_evaluated, metrics, relative_step)
        self._update_logger(trainer, task_evaluated, metrics, global_step)
        self._update_tuner(is_final_score=False)
    
    def _collect_eval_metrics(self, trainer, start_task: int, end_task: int):
        global_step = trainer.current_task if self.eval_interval == 'task' else trainer.current_epoch
        relative_step = self.interval_calculator.get_step_within_task(trainer.current_epoch)
        for eval_task in range(start_task, end_task + 1):
            task_metrics = trainer.validate_algorithm_on_task(eval_task)
            self.log_metrics(trainer, trainer.current_task, eval_task, task_metrics, global_step, relative_step)
            print(f"[{global_step}] Eval metrics for task {eval_task} >> {task_metrics}")

    def save_metrics(self):
        metrics = ['accuracy', 'loss'] if self.eval_type == 'classification' else ['loss']
        for metric in metrics:
            filepath = os.path.join(self.save_paths['metrics'], metric + ".npy")
            with open(filepath, 'wb') as f:
                np.save(f, self.meters[metric].data)
    
    def _prepare_plot_params(self):
        xticks = self.interval_calculator.get_tick_times()
        if self.collect_on_init:
            xticks = [0] + xticks
        return {
            'show_legend': True,
            'legend_loc': 'lower left',
            'xticks': xticks,
            'xlabel': 'Epochs',
            'ylabel': 'Validation Accuracy'
        }
    
    def _extract_task_history(self, task):
        if self.collect_on_init:
            start, end = 0, self.interval_calculator.get_task_range(task)[1]
            offset = 0 if self.eval_interval == 'task' else self.epochs_per_task - 1
            metrics = self.meters['accuracy'].get_raw_history(task, 0)[start+offset:end+offset]
            # print(metrics)
        elif self.collect_metrics_for_future_tasks:
            start, end = 1, self.interval_calculator.get_task_range(task)[1]
            metrics = self.meters['accuracy'].get_raw_history(task, 1)
        else:
            start, end = self.interval_calculator.get_task_range(task)
            metrics = self.meters['accuracy'].get_raw_history(task)[start-1:]
            if self.eval_interval == 'task':
                metrics = self.meters['accuracy'].get_raw_history(task)
                start = self.epochs_per_task-1
                metrics = metrics[start::self.epochs_per_task]
                print(f"DEBUG >> task {task}, metric shape= {metrics.shape}")
        return range(start, end), metrics
        
    def plot_metrics(self, logger: Optional[Logger] = None):
        if self.eval_type != 'classification':
            return
        
        plot_params = self._prepare_plot_params()
        data, labels = [], []
        for task in range(1, self.num_tasks + 1):
            data.append(self._extract_task_history(task))
            labels.append(f"Task{task}")
        Visualizer.line_plot(data, labels=labels, save_dir=self.save_paths['plots'], filename="metrics",
                             cmap='g10', plot_params=plot_params, logger=logger)
    
    def on_before_fit(self, trainer):
        if self.collect_on_init:
            print(f"---------------------------- Init -----------------------")
            self._collect_eval_metrics(trainer, 1, end_task=self.num_tasks)
    
    def on_after_training_epoch(self, trainer):
        if self.eval_interval != 'epoch':
            return
        if self.collect_on_init or self.collect_metrics_for_future_tasks:
            self._collect_eval_metrics(trainer, 1, self.num_tasks)
        else:
            self._collect_eval_metrics(trainer, 1, trainer.current_task)
    
    def on_before_training_task(self, trainer):
        print(f"---------------------------- Task {trainer.current_task+1} -----------------------")
        
    def on_after_training_task(self, trainer):
        if self.eval_interval != 'task':
            return
        if self.collect_on_init or self.collect_metrics_for_future_tasks:
            self._collect_eval_metrics(trainer, 1, self.num_tasks)
        else:
            self._collect_eval_metrics(trainer, 1, trainer.current_task)

    def on_after_fit(self, trainer):
        self._update_tuner(is_final_score=True)
        self.save_metrics()
        self.plot_metrics(trainer.logger)


import torchvision
from typing import Optional, Tuple
from cl_gym.benchmarks.utils import DEFAULT_DATASET_DIR
from cl_gym.benchmarks.base import Benchmark, DynamicTransformDataset
from cl_gym.benchmarks.transforms import get_default_mnist_fashion_mnist_transform


class MNISTFashionMNIST(Benchmark):
    def __init__(self,
                 num_tasks: int = 2,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):
        
        if num_tasks != 2:
            raise ValueError("MNIST-FashionMNIST benchmark can have only two tasks")
        
        if task_input_transforms is None:
            task_input_transforms = get_default_mnist_fashion_mnist_transform()
            
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms)
        
        self.load_datasets()
        self.prepare_datasets()
    
    def __load_mnist(self):
        self.mnist_train = torchvision.datasets.MNIST(DEFAULT_DATASET_DIR, train=True, download=True)
        self.mnist_test = torchvision.datasets.MNIST(DEFAULT_DATASET_DIR, train=False, download=True)
        
    def __load_fashion_mnist(self):
        self.mnist_train = torchvision.datasets.FashionMNIST(DEFAULT_DATASET_DIR, train=True, download=True)
        self.mnist_test = torchvision.datasets.FashionMNIST(DEFAULT_DATASET_DIR, train=False, download=True)

    def load_datasets(self):
        for task in range(1, self.num_tasks + 1):
            if task == 1:
                self.__load_mnist()
            else:
                self.__load_fashion_mnist()
            input_transform = self.task_input_transforms[task - 1]
            target_transform = self.task_target_transforms[task - 1] if self.task_target_transforms else None
            self.trains[task] = DynamicTransformDataset(task, self.mnist_train, input_transform, target_transform)
            self.tests[task] = DynamicTransformDataset(task, self.mnist_test, input_transform, target_transform)
    
    def precompute_memory_indices(self):
        for task in range(1, self.num_tasks + 1):
            indices_train = self.sample_uniform_class_indices(self.trains[task].dataset, 0, 9,
                                                              self.per_task_memory_examples)
            indices_test = self.sample_uniform_class_indices(self.tests[task].dataset, 0, 9,
                                                             self.per_task_memory_examples)
            assert len(indices_train) == len(indices_test) == self.per_task_memory_examples
            self.memory_indices_train[task] = indices_train[:]
            self.memory_indices_test[task] = indices_test[:]




from cl_gym.utils.metrics.base import ContinualMetric
from cl_gym.utils.metrics.accuracy import PerformanceMetric
from cl_gym.utils.metrics.forgetting import ForgettingMetric

__all__ = ['ContinualMetric',
           'PerformanceMetric',
           'ForgettingMetric']



import torchvision
from typing import Optional, Tuple
from cl_gym.benchmarks.utils import DEFAULT_DATASET_DIR
from cl_gym.benchmarks.transforms import get_default_mnist_transform
from cl_gym.benchmarks.transforms import get_default_rotation_mnist_transform
from cl_gym.benchmarks.transforms import get_default_permuted_mnist_transform
from cl_gym.benchmarks.base import Benchmark, DynamicTransformDataset, SplitDataset


class ContinualMNIST(Benchmark):
    """
    Base class for (Permuted/Rotated/Split)-MNIST benchmarks.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):
        """
        Args:
            num_tasks: The number of tasks for the benchmark.
            per_task_examples: If set, each task will include part of the original benchmark rather than full data.
            per_task_joint_examples: If set, the benchmark will support joint/multitask loading of tasks.
            per_task_memory_examples: If set, the benchmark will support episodic memory/replay buffer loading of tasks.
            per_task_subset_examples: If set, the benchmark will support loading a pre-defined subset of each task.
            task_input_transforms: If set, the benchmark will use the provided torchvision transform.
            task_target_transforms: If set, the benchmark will use the provided target transform for targets.
            
        . note::
            If :attr:`task_input_transforms` or :attr:`task_target_transforms`, they should be a list
            of size `num_tasks` where each element of the list can be a torchvision (Composed) transform.
        """
        
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms)

        self.load_datasets()
        self.prepare_datasets()

    def __load_mnist(self):
        self.mnist_train = torchvision.datasets.MNIST(DEFAULT_DATASET_DIR, train=True, download=True)
        self.mnist_test = torchvision.datasets.MNIST(DEFAULT_DATASET_DIR, train=False, download=True)
    
    def load_datasets(self):
        self.__load_mnist()
        for task in range(1, self.num_tasks+1):
            input_transform = self.task_input_transforms[task-1]
            target_transform = self.task_target_transforms[task-1] if self.task_target_transforms else None
            self.trains[task] = DynamicTransformDataset(task, self.mnist_train, input_transform, target_transform)
            self.tests[task] = DynamicTransformDataset(task, self.mnist_test, input_transform, target_transform)
    
    def precompute_memory_indices(self):
        for task in range(1, self.num_tasks + 1):
            indices_train = self.sample_uniform_class_indices(self.trains[task].dataset, 0, 9, self.per_task_memory_examples)
            indices_test = self.sample_uniform_class_indices(self.tests[task].dataset, 0, 9, self.per_task_memory_examples)
            assert len(indices_train) == len(indices_test) == self.per_task_memory_examples
            self.memory_indices_train[task] = indices_train[:]
            self.memory_indices_test[task] = indices_test[:]
            

class RotatedMNIST(ContinualMNIST):
    """
    Rotated MNIST benchmark.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None,
                 per_task_rotation: Optional[float] = None):
        """
        Args:
            num_tasks: The number of tasks for the benchmark.
            per_task_examples: If set, each task will include part of the original benchmark rather than full data.
            per_task_joint_examples: If set, the benchmark will support joint/multitask loading of tasks.
            per_task_memory_examples: If set, the benchmark will support episodic memory/replay buffer loading of tasks.
            per_task_subset_examples: If set, the benchmark will support loading a pre-defined subset of each task.
            task_input_transforms: If set, the benchmark will use the provided torchvision transform.
            task_target_transforms: If set, the benchmark will use the provided target transform for targets.
            per_task_rotation: rotation degrees per task.
        
        . note::
            The first task won't have any rotations.
            e.g., if rotation per task is 45 degrees, first task has 0 deg rotation, second task has 45 degree, etc.
        """
        
        if task_input_transforms is None:
            task_input_transforms = get_default_rotation_mnist_transform(num_tasks, per_task_rotation)
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms)


class PermutedMNIST(ContinualMNIST):
    """
    Permuted MNIST benchmark
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):
        """
        Args:
            num_tasks: The number of tasks for the benchmark.
            per_task_examples: If set, each task will include part of the original benchmark rather than full data.
            per_task_joint_examples: If set, the benchmark will support joint/multitask loading of tasks.
            per_task_memory_examples: If set, the benchmark will support episodic memory/replay buffer loading of tasks.
            per_task_subset_examples: If set, the benchmark will support loading a pre-defined subset of each task.
            task_input_transforms: If set, the benchmark will use the provided torchvision transform.
            task_target_transforms: If set, the benchmark will use the provided target transform for targets.
        
        . note::
            The first task will be the default MNIST and the permutation applies to tasks 2, 3, ..., `num_tasks`.
        """
        if task_input_transforms is None:
            task_input_transforms = get_default_permuted_mnist_transform(num_tasks)
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms)


class SplitMNIST(ContinualMNIST):
    """
    Split MNIST benchmark.
    The benchmark can have at most 5 tasks, each a binary classification on MNIST digits.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):
        if num_tasks > 5:
            raise ValueError("Split MNIST benchmark can have at most 5 tasks (i.e., 10 classes, 2 per task)")
        if task_input_transforms is None:
            task_input_transforms = get_default_mnist_transform(num_tasks)
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms)

    def __load_mnist(self):
        transforms = self.task_input_transforms[0]
        self.mnist_train = torchvision.datasets.MNIST(DEFAULT_DATASET_DIR, train=True, download=True, transform=transforms)
        self.mnist_test = torchvision.datasets.MNIST(DEFAULT_DATASET_DIR, train=False, download=True, transform=transforms)

    def load_datasets(self):
        self.__load_mnist()
        for task in range(1, self.num_tasks + 1):
            self.trains[task] = SplitDataset(task, 2, self.mnist_train)
            self.tests[task] = SplitDataset(task, 2, self.mnist_test)

    def precompute_memory_indices(self):
        for task in range(1, self.num_tasks + 1):
            start_cls = (task - 1) * 2
            end_cls = task * 2 - 1
            num_examples = self.per_task_memory_examples
            indices_train = self.sample_uniform_class_indices(self.trains[task], start_cls, end_cls, num_examples)
            indices_test = self.sample_uniform_class_indices(self.tests[task], start_cls, end_cls, num_examples)
            assert len(indices_train) == len(indices_test) == self.per_task_memory_examples
            self.memory_indices_train[task] = indices_train[:]
            self.memory_indices_test[task] = indices_test[:]



from cl_gym.utils.callbacks import ContinualCallback


class ExperimentManager(ContinualCallback):
    """
    Experiment manager callback: logs parameters before training starts, updates server when training ends.
    """
    def __init__(self):
        super(ExperimentManager, self).__init__('ExperimentManager')
    
    def on_before_fit(self, trainer):
        if trainer.logger:
            trainer.logger.log_parameters(trainer.params)
    
    def on_before_teardown(self, trainer):
        if trainer.logger:
            path = trainer.params['output_dir']
            trainer.logger.log_folder(folder_path=path)
            trainer.logger.terminate()



import torch
import numpy as np
from typing import Optional
from torch.utils.data import Dataset
from cl_gym.benchmarks import Benchmark

DEFAULT_TOY_DATASET_SIZE = 100
DEFAULT_STD = 0.05


class Toy1DRegDataset(Dataset):
    """
    Toy task: 1D Regression problem with 3 tasks
    This task was introduced by Oswald & Henning et. al.
    Please see: https://openreview.net/pdf?id=SJgwNerKvB
    """
    def __init__(self, task_id, num_examples: int, noise_std: float = 0.05):
        self.task_id = task_id
        self.num_examples = num_examples
        self.noise_std = noise_std
        self.data = []
        self.targets = []
        self.__generate_data()
    
    def __generate_data(self):
        map_functions = [lambda x: (x + 3.),
                     lambda x: 2. * np.power(x, 2) - 1,
                     lambda x: np.power(x - 3., 3)]
        x_domains = [[-4, -2], [-1, 1], [2, 4]]
        
        start, end = x_domains[self.task_id-1]
        self.data = np.random.uniform(low=start, high=end, size=(self.num_examples, 1))
        self.targets = map_functions[self.task_id-1](self.data)
        if self.noise_std > 0.0:
            noise = np.random.normal(loc=0.0, scale=self.noise_std, size=(self.num_examples, 1))
            self.targets += noise
        
        assert len(self.data) == len(self.targets)
        self.data = torch.from_numpy(self.data).float()
        self.targets = torch.from_numpy(self.targets).float()

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index: int):
        return self.data[index], self.targets[index], self.task_id


class Toy1DRegression(Benchmark):
    """
    Toy benchmark: 1D Regression problem with 3 tasks
    Task `t` will be a polynomial with degree `t`
    Please see: https://openreview.net/pdf?id=SJgwNerKvB
    """
    def __init__(self,
                 num_tasks: int = 3,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_size: Optional[int] = DEFAULT_TOY_DATASET_SIZE,
                 noise_std: Optional[float] = DEFAULT_STD):
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples)
        self.task_size = task_size
        self.noise_std = noise_std
        self.load_datasets()
        self.prepare_datasets()
    
    def load_datasets(self):
        for task in range(1, self.num_tasks + 1):
            self.trains[task] = Toy1DRegDataset(task, self.task_size, self.noise_std)
            self.tests[task] = Toy1DRegDataset(task, self.task_size, self.noise_std)
    
    def precompute_memory_indices(self):
        for task in range(1, self.num_tasks + 1):
            self.memory_indices_train[task] = np.random.randint(0, len(self.trains[task]), size=self.per_task_memory_examples)
            self.memory_indices_test[task] = np.random.randint(0, len(self.tests[task]), size=self.per_task_memory_examples)



import torch
from torch import nn
from torch import optim
from torch.nn import functional as F
from torch import autograd
from cl_gym.algorithms import ContinualAlgorithm


class EWC(ContinualAlgorithm):
    """
    | Elastic Weight Consolidation
    | By Kirkpatricka et al. : https://arxiv.org/abs/1612.00796.pdf
    """
    # implementation is partially based on: https://github.com/kuc2477/pytorch-ewc
    def __init__(self, backbone, benchmark, params, fisher_lambda: float = 50.0, fisher_sample_size: int = 128):
        """
        Args:
            backbone: the backbone model
            benchmark: the benchmark
            params: params for training
            fisher_lambda: The lambda coefficient of EWC algorithm
            fisher_sample_size: Sample size for calculating Fisher diagonal
        """
        super(EWC, self).__init__(backbone, benchmark, params)
        self.fisher_lambda = fisher_lambda
        self.fisher_sample_size = fisher_sample_size
    
    def __estimate_fisher_diagonal(self):
        log_likelihoods = []
        samples_so_far = 0
        train_loader, _ = self.benchmark.load(self.current_task, batch_size=32)
        for x, y, _ in train_loader:
            batch_size = len(y)
            x = x.to(self.params['device'])
            y = y.to(self.params['device'])
            log_out = F.log_softmax(self.backbone(x, self.current_task), dim=1)
            log_likelihoods.append(log_out[range(batch_size), y.data])
            samples_so_far += batch_size
            if samples_so_far > self.fisher_sample_size:
                break
        
        log_likelihoods = torch.cat(log_likelihoods).unbind()
        grads = zip(*[autograd.grad(l, self.backbone.parameters(), retain_graph=(i < len(log_likelihoods)))\
                      for i, l in enumerate(log_likelihoods, 1)])
        grads = [torch.stack(grad) for grad in grads]
        fisher_diagonals = [(grad ** 2).mean(0) for grad in grads]
        
        # note for the next line: in pytorch, module names are like W1.weight
        # but, we can't get attrs using getattr('W1.weight') because of the nested call (dot)
        # one trick is to replace the '.' with '_'
        # the other tick is to use: functools.reduce(getattr, [obj] + attr.split('.'))
        param_names = [n.replace('.', '_') for n, p in self.backbone.named_parameters()]
        return {n: f.detach() for n, f in zip(param_names, fisher_diagonals)}
    
    def __consolidate(self):
        fisher_diagonals = self.__estimate_fisher_diagonal()
        for name, param in self.backbone.named_parameters():
            name = name.replace('.', '_')
            self.backbone.register_buffer(f"{name}_mean", param.data.clone())
            self.backbone.register_buffer(f"{name}_fisher", fisher_diagonals[name].data.clone())
    
    def __calculate_ewc_loss(self):
        # shouldn't be called for the first task
        # because we have not consolidated anything yet
        losses = []
        for name, param in self.backbone.named_parameters():
            name = name.replace('.', '_')
            mean = getattr(self.backbone, f"{name}_mean")
            fisher = getattr(self.backbone, f"{name}_fisher")
            losses.append((fisher * (param - mean)**2).sum())
        
        return (self.fisher_lambda/2.0)*sum(losses)
    
    def training_task_end(self):
        self.__consolidate()
        self.current_task += 1
        
    def training_step(self, task_id, inp, targ, optimizer, criterion):
        optimizer.zero_grad()
        pred = self.backbone(inp, task_id)
        loss = criterion(pred, targ)
        if task_id > 1:
            loss += self.__calculate_ewc_loss()
        loss.backward()
        optimizer.step()


import torch
import torch.nn as nn
import torch.nn.functional as F


class CNN1D(nn.Module):
    """
    A small 1-D CNN used for time series tasks.
    """
    def __init__(self, dropout_prob=0.0):
        super().__init__()
        self.dropout_prob = dropout_prob
        self.conv1 = nn.Conv1d(18, 64, 5)
        self.conv2 = nn.Conv1d(64, 64, 5)
        self.dropout = nn.Dropout(p=dropout_prob)
        self.pool2 = nn.MaxPool1d(4)
        self.relu = nn.ReLU(inplace=True)
        self.fc3 = nn.Linear(64*62, 200)
        self.fc4 = nn.Linear(200, 12)

    def forward(self, x, task_id=None):
        out = x
        batch_size = out.shape[0]
        out = torch.transpose(out, 1, 2)
        out = self.conv1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.relu(out)
        if self.dropout_prob:
            out = self.dropout(out)
        out = self.pool2(out)
        out = out.view(batch_size, -1)
        out = self.fc3(out)
        out = self.relu(out)
        out = self.fc4(out)
        return out


from cl_gym.backbones.base import ContinualBackbone
from cl_gym.backbones.mlp import MLP2Layers
from cl_gym.backbones.cnn import CNN1D
from cl_gym.backbones.resnet import ResNet18Small

__all__ = ['ContinualBackbone', 'MLP2Layers', 'CNN1D', 'ResNet18Small']



import os
import copy
import torch
import matplotlib
import numpy as np
import seaborn as sns
from pathlib import Path
import matplotlib.pylab as pl
from matplotlib import pyplot as plt
from typing import List, Tuple, Union, Iterable, Optional
from cl_gym.utils.loggers import Logger


class IntervalCalculator:
    def __init__(self, num_tasks: int,
                 epochs_per_task: int,
                 interval: str):
        # interval can be either 'epoch' or 'task'
        self.num_tasks = num_tasks
        self.epochs_per_task = epochs_per_task
        self.interval = interval
    
    def get_task_range(self, target_task: int):
        if self.interval == 'task':
            start, end = target_task, self.num_tasks + 1
        elif self.interval == 'epoch':
            start = (target_task - 1) * self.epochs_per_task + 1
            end = self.num_tasks * self.epochs_per_task + 1
        else:
            raise ValueError("Supported intervals are `epoch` and `task`")
        return start, end
    
    def get_tick_times(self):
        if self.interval == 'task':
            return range(1, self.num_tasks+1)
        elif self.interval == 'epoch':
            return list(range(1, (self.num_tasks+1)*self.epochs_per_task, self.epochs_per_task))[:-1] \
                   + [self.num_tasks*self.epochs_per_task]
        else:
            raise ValueError("Supported intervals are `epoch` and `task`")
    
    def get_step_within_task(self, global_epoch: int):
        if self.interval == 'task':  # task-based interval: only one epoch
            return 1
        else:  # epoch-based interval
            relative_epoch = global_epoch % self.epochs_per_task
            if relative_epoch == 0:  # last epoch of the task
                return self.epochs_per_task
            return relative_epoch


class Visualizer:
    def __init__(self):
        pass
    
    @staticmethod
    def get_pallete(palette_name='str'):
        palletes = {
            'plotly': ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A',
                       '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52'],
            'tensorboard': ['#ff7043', '#33bbee', '#ee3377', '#009988', '#0077bb',
                            '#cc3311', '#7043ff', "#FEB843", "#bbbbbb", '#120031'],
            't10': ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b',
                    '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'],
            'g10': ['#3366CC', '#FF9900', '#109618', '#DC3912', '#990099', '#0099C6',
                    '#DD4477', '#66AA00', '#B82E2E', '#316395'],
            'colorblind': ["#0072B2", "#D55E00", "#E0D53B", "#009E73", "#C244E1",
                           '#56B4E9', '#a6761d', '#cc001d', '#00cc49', '#000000'],
            'bright': ['#0053d6', '#e1144b', '#e1aa14', '#00d783', '#aa14e1',
                       '#ff630a', '#00bed6', '#bed600', '#d600be', '#120031']}
        return palletes[palette_name.lower()]
    
    
    @staticmethod
    def generate_colors(num_colors: int):
        colors = pl.cm.jet(np.linspace(0, 1, num_colors))
        return colors

    @staticmethod
    def reset():
        plt.close('all')
    
    @staticmethod
    def set_context():
        rc = {"lines.linewidth": 3.5,
              'xtick.labelsize': 18,
              'ytick.labelsize': 18,
              'lines.markersize': 9,
              'axes.labelsize': 20,
              'axes.titlesize': 22,
              'legend.title_fontsize': 18,
              'legend.fontsize': 16,
              'legend.handlelength': 1,
              'legend.handleheight': 1,
              'text.usetex': False}
        
        sns.set_context("notebook", rc=rc)
    
    @staticmethod
    def _prepare_plot_context(plot_type: str, rc_dict=None):
        # plot_type supports 'line' and 'heatmap'
        Visualizer.set_context()
        if rc_dict:
            plt.rcParams.update(rc_dict)
        
        if plot_type.lower() == 'line':
            sns.set_style('whitegrid')
        elif plot_type.lower() == 'heatmap':
            sns.set_style('white')
    
    @staticmethod
    def _apply_plot_details(plot_params):
        if plot_params:
            if plot_params.get('title'):
                plt.title(plot_params['title'])
            if plot_params.get('xlabel'):
                plt.xlabel(plot_params['xlabel'])
            if plot_params.get('ylabel'):
                plt.ylabel(plot_params['ylabel'])
            if plot_params.get("xlim"):
                plt.xlim(plot_params['xlim'])
            if plot_params.get("ylim"):
                plt.ylim(plot_params['ylim'])
            if plot_params.get("xticks"):
                if isinstance(plot_params['xticks'], list):
                    plt.xticks(plot_params['xticks'])
                elif isinstance(plot_params['xticks'], dict):
                    plt.xticks(plot_params['xticks']['ticks'], plot_params['xticks']['labels'])
            if plot_params.get("yticks"):
                if isinstance(plot_params['yticks'], list):
                    plt.yticks(plot_params['yticks'])
                elif isinstance(plot_params['yticks'], dict):
                    plt.xticks(plot_params['yticks']['ticks'], plot_params['yticks']['labels'])
            if plot_params.get("show_legend"):
                legend_loc = plot_params.get("legend_loc", "best")
                legend_ncol = plot_params.get("legend_ncol", 1)
                plt.legend(loc=legend_loc, ncol=legend_ncol)
        
            if plot_params.get("tight_layout"):
                plt.tight_layout()
    
    @staticmethod
    def line_plot(lines: List[Tuple[Iterable, np.ndarray]],
                  lines_std: List[Tuple[Iterable, np.ndarray]] = None,
                  labels: Optional[List[str]] = None,
                  colors: Optional[Union[List[str]]] = None,
                  cmap: Optional[str] = 'colorblind',
                  rc_params: Optional[dict] = None,
                  plot_params: Optional[dict] = None,
                  logger: Optional[Logger] = None,
                  logger_step: Optional[int] = None,
                  save_dir: Optional[str] = None,
                  filename: Optional[str] = None,
                  save_format: Optional[str] = 'pdf'):
        
        # setup
        Visualizer.reset()
        Visualizer._prepare_plot_context('line', rc_params)
        
        if not colors and cmap and len(lines) <= 10:
            colors = Visualizer.get_pallete(cmap)
        else:
            colors = Visualizer.generate_colors(len(lines))
        # plot lines
        for i, line in enumerate(lines):
            x, y = line
            color = colors[i]
            label = labels[i] if labels else None
            plt.plot(x, y, color=color, label=label)
            if lines_std:
                plt.fill_between(x, y-lines_std[i], y+lines_std, color=color, alpha=0.4)
        
        # plot details
        Visualizer._apply_plot_details(plot_params)
        
        if logger:
            logger.log_figure(plt, filename, logger_step)
        
        plt.savefig(os.path.join(save_dir, f"{filename}.{save_format}"),
                    dpi=200, format=save_format, bbox_inches='tight')
        plt.close('all')


import comet_ml
from comet_ml import Experiment
import os


class Logger:
    def __init__(self, api_key, project_name, trial_name):
        self.api_key = api_key
        self.trial_id = trial_name
        self.project_name = project_name
        self.experiment = self._build_experiment()
    
    def _build_experiment(self):
        raise NotImplementedError
    
    def log_parameters(self, params: dict):
        raise NotImplementedError
    
    def log_metric(self, key: str, value, step=None):
        raise NotImplementedError
    
    def log_figure(self, figure, name, step=None):
        raise NotImplementedError
    
    def log_text(self, text):
        raise NotImplementedError
    
    def log_folder(self, folder_path):
        raise NotImplementedError
    
    def terminate(self):
        raise NotImplementedError


class CometLogger(Logger):
    def __init__(self, api_key=None, project_name=None, trial_name=None, workspace=None):
        if not api_key:
            api_key = os.environ.get("LOGGER_API_KEY")
            if not api_key:
                raise ValueError("Either set `LOGGER_API_KEY` environment variable or pass the API key to the class")
        if not project_name or not workspace:
            raise ValueError("Project & workspace name for the comet experiment manager is required.")
        self.workspace = workspace
        super(CometLogger, self).__init__(api_key, project_name, trial_name)
    
    def _build_experiment(self):
        exp = Experiment(self.api_key, self.project_name, self.workspace,
                          auto_metric_logging=False, auto_param_logging=False,
                          log_graph=False, disabled=False)
        exp.disable_mp()
        return exp
    
    def log_parameters(self, params: dict):
        self.experiment.log_parameters(params)
    
    def log_metric(self, key: str, value, step=None):
        self.experiment.log_metric(key, value, step)
    
    def log_text(self, text):
        self.experiment.log_text(text)
    
    def log_figure(self, figure, name, step=None):
        self.experiment.log_figure(name, figure, step)
    
    def log_folder(self, folder_path):
        self.experiment.log_asset_folder(folder_path, log_file_name=False, recursive=True)
    
    def terminate(self):
        self.experiment.end()



import torch
from torch import nn
from torch import optim
from cl_gym.algorithms import ContinualAlgorithm


class Multitask(ContinualAlgorithm):
    """
    Multitask (Joint) Training
    """
    def __init__(self, backbone, benchmark, params):
        super(Multitask, self).__init__(backbone, benchmark, params, requires_memory=True)

    def prepare_train_loader(self, task_id):
        return self.benchmark.load_joint(task_id, self.params['batch_size_train'], shuffle=True, pin_memory=True)[0]



import torchvision
from typing import Optional, Tuple
from cl_gym.benchmarks.utils import DEFAULT_DATASET_DIR
from cl_gym.benchmarks import Benchmark, SplitDataset
from cl_gym.benchmarks.transforms import get_default_cifar_transform


class SplitCIFAR(Benchmark):
    """
    Base class for Split-CIFAR benchmarks.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None,
                 is_cifar_100: bool = True):
        """
        Args:
            num_tasks: Number of tasks. 20 for CIFAR-100 and 5 for CIFAR-10.
            per_task_examples: If set, each task will include part of the original benchmark rather than full data.
            per_task_joint_examples: If set, the benchmark will support joint/multitask loading of tasks.
            per_task_memory_examples: If set, the benchmark will support episodic memory/replay buffer loading of tasks.
            per_task_subset_examples: If set, the benchmark will support loading a pre-defined subset of each task.
            task_input_transforms: If set, the benchmark will use the provided torchvision transform.
            task_target_transforms: If set, the benchmark will use the provided target transform for targets.
            is_cifar_100: If true, it will set prepare for CIFAR-100, otherwise CIFAR-10.
            
        . note::
            If :attr:`task_input_transforms` or :attr:`task_target_transforms`, they should be a list
            of size `num_tasks` where each element of the list can be a torchvision (Composed) transform.
            
        . note::
            Similar to Torchvision, CIFAR benchmarks in CL-Gym are first loaded in the memory for faster loading.
        """
        
        # CIFAR-100 vs CIFAR-10 book-keeping variables:
        # CIFAR-100 has 20 tasks (5 classes per task)
        # CIFAR-10  has 5 tasks (2 classes per task)
        self.is_cifar_100 = is_cifar_100
        if task_input_transforms is None:
            task_input_transforms = get_default_cifar_transform(num_tasks, is_cifar_100)
        self.num_classes_per_split = 5 if self.is_cifar_100 else 2
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms)
        self.load_datasets()
        self.prepare_datasets()

    def __load_cifar(self):
        transforms = self.task_input_transforms[0]
        CIFAR_dataset = torchvision.datasets.CIFAR100 if self.is_cifar_100 else torchvision.datasets.CIFAR10
        self.cifar_train = CIFAR_dataset(DEFAULT_DATASET_DIR, train=True, download=True, transform=transforms)
        self.cifar_test = CIFAR_dataset(DEFAULT_DATASET_DIR, train=False, download=True, transform=transforms)

    def load_datasets(self):
        """
        Loades CIFAR-Dataset [In memory]
        """
        self.__load_cifar()
        for task in range(1, self.num_tasks + 1):
            self.trains[task] = SplitDataset(task, self.num_classes_per_split, self.cifar_train)
            self.tests[task] = SplitDataset(task, self.num_classes_per_split, self.cifar_test)
    
    def precompute_memory_indices(self):
        """
        Precomputes memory indices for each task.
        
        . note::
            The default behavior is class-uniform sampling.
            i.e., each class will have roughly equal number of samples in the memory.
            You can inherit this class and override this method for custom behavior. But a better way
            is to move this logic to your algorithm component's code.
        """
        for task in range(1, self.num_tasks + 1):
            start_cls = (task - 1) * self.num_classes_per_split
            end_cls = task * self.num_classes_per_split - 1
            num_examples = self.per_task_memory_examples
            indices_train = self.sample_uniform_class_indices(self.trains[task], start_cls, end_cls, num_examples)
            indices_test = self.sample_uniform_class_indices(self.tests[task], start_cls, end_cls, num_examples)
            assert len(indices_train)  == self.per_task_memory_examples
            self.memory_indices_train[task] = indices_train[:]
            self.memory_indices_test[task] = indices_test[:]


class SplitCIFAR100(SplitCIFAR):
    """
    Split CIFAR-100 benchmark.
    Has 20 tasks, each task with 5 classes of CIFAR-100.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):

        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms, is_cifar_100=True)


class SplitCIFAR10(SplitCIFAR):
    """
    Split CIFAR-10 benchmark.
    has 5 tasks, each with 2 classes of CIFAR-10.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 task_input_transforms: Optional[list] = None,
                 task_target_transforms: Optional[list] = None):
        
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples, task_input_transforms, task_target_transforms, is_cifar_100=False)




import torch
import torch.nn as nn
from typing import Optional, Dict, Iterable
from cl_gym.backbones import ContinualBackbone

activations = {
    'relu': nn.ReLU(),
    'tanh': nn.Tanh(),
    'sigmoid': nn.Sigmoid(),
}


class FCBlock(nn.Module):
    """
    Fully-Connected block with activations
    (i.e., linear layer, followed by ReLU and [optionally] a dropout layer)
    """
    def __init__(self, inp_dim: int, out_dim: int,
                 dropout_prob: float = 0.0,
                 include_activation: bool = True,
                 bias: bool = True,
                 activation: str = 'relu'):
        """
        
        Args:
            inp_dim: Input dimension.
            out_dim: Output dimension.
            dropout_prob: Dropout probability. Default: 0.0 (No dropout)
            include_activation: Should a linear layer followed by an activation layer?
            bias: Should a linear layer have bias?
            activation: Activation function. Currently supports `relu`, `tanh`, `sigmoid`.
        """
        
        super(FCBlock, self).__init__()
        layers = [nn.Linear(inp_dim, out_dim, bias=bias)]
        if include_activation:
            layers.append(activations[activation.lower()])
        if dropout_prob > 0.0:
            self.layers.append(nn.Dropout(dropout_prob))
        self.layers = nn.ModuleList(layers)
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = x
        for layer in self.layers:
            out = layer(out)
        return out


class MLP2Layers(ContinualBackbone):
    """
    MLP model (feed-forward) with two hidden layers.
    """
    def __init__(self, multi_head=False, num_classes_per_head=None,
                 input_dim=784, hidden_dim_1=256, hidden_dim_2=256,
                 output_dim=10, dropout_prob=0.0, activation='ReLU',
                 bias=True, include_final_layer_act=False):
        """
        
        Args:
            multi_head: Is this a multi-head model?
            num_classes_per_head: If backbone is multi-head, then what is the head size?
            input_dim:
            hidden_dim_1:
            hidden_dim_2:
            output_dim:
            dropout_prob: Dropout probability. Default: No dropout.
            activation: The name of activation function. Default: 'ReLU'.
            bias: Should linear layers have bias?
            include_final_layer_act: Should the last layer have activation function? Default: False.
        """
        # model variables
        super(MLP2Layers, self).__init__(multi_head, num_classes_per_head)
        self.input_dim = input_dim
        self.hidden_dim_1 = hidden_dim_1
        self.hidden_dim_2 = hidden_dim_2
        self.output_dim = output_dim
        self.dropout_prob = dropout_prob
        block_1 = FCBlock(self.input_dim, self.hidden_dim_1, self.dropout_prob, True, bias, activation)
        block_2 = FCBlock(self.hidden_dim_1, self.hidden_dim_2, self.dropout_prob, True, bias, activation)
        block_3 = FCBlock(self.hidden_dim_2, self.output_dim, 0.0, include_final_layer_act, bias, activation)
        self.blocks: nn.ModuleList = nn.ModuleList([block_1, block_2, block_3])
    
    @torch.no_grad()
    def get_block_params(self, block_id: int) -> Dict[str, torch.Tensor]:
        """
        Args:
            block_id: the block number. In this case, layer.

        Returns:
            params: a dictionary of form {'weight': weight_params, 'bias': bias_params}
        """
        assert 1 <= block_id <= 3
        block = self.blocks[block_id-1].layers[0]
        weight = block.weight.data if block.weight is not None else None
        bias = block.bias.data if block.bias is not None else None
        return {'weight': weight, 'bias': bias}
    
    @torch.no_grad()
    def get_block_outputs(self, inp: torch.Tensor, block_id: int, pre_act: bool = False):
        assert 1 <= block_id <= 3
        out = inp.view(inp.shape[0], -1)
        current_block = 0
        while current_block < block_id:
            out = self.blocks[current_block].layers[0](out)
            if not pre_act and len(self.blocks[current_block].layers) > 1:
                out = self.blocks[current_block].layers[1](out)
            current_block += 1
        return out.detach()
    
    def get_block_grads(self, block_id: int) -> Dict[str, Optional[torch.Tensor]]:
        block = self.blocks[block_id-1].layers[0]
        weight_grad, bias_grad = None, None
        if block.weight is not None and block.weight.grad is not None:
            weight_grad = block.weight.grad.data
        if block.bias is not None and block.bias.grad is not None:
            bias_grad = block.bias.grad.data
        return {'weight': weight_grad, 'bias': bias_grad}

    def forward(self, inp: torch.Tensor, head_ids: Optional[Iterable] = None) -> torch.Tensor:
        """
        Args:
            inp: The input of shape [BatchSize x input_dim]
            head_ids: Optional iterable (e.g., List or 1-D Tensor) object of shape [BatchSize] includes head_ids.

        Returns:
            output: The forward-pass output. Shape: [BatchSize x output_dim]
        
        Note: the `head_id` will only be used if the backbone is initiated with `multi_head = True`.
        """
        inp = inp.view(inp.shape[0], -1)
        return super(MLP2Layers, self).forward(inp, head_ids)



import numpy as np
from typing import Optional
from cl_gym.utils.metrics import ContinualMetric


class ForgettingMetric(ContinualMetric):
    def __init__(self, num_tasks: int, epochs_per_task: Optional[int] = 1):
        super().__init__(num_tasks, epochs_per_task)
    
    def compute(self, current_task) -> float:
        # forgetting is defined for more than one task!
        if current_task == 1:
            return 0.0
        total_forgetting = 0
        for task in range(1, current_task):
            max_acc = np.max(self.data[1:, task, -1])
            last_acc = self.data[current_task][task][-1]
            task_forgetting = max_acc - last_acc
            total_forgetting += task_forgetting
        return total_forgetting/(current_task-1)

    def compute_final(self) -> float:
        return self.compute(self.num_tasks)

# if __name__ == "__main__":
#     avg_acc = AverageForgetting(3, 2)
#     # init
#     avg_acc.update(0, 1, 50, 1)
#     avg_acc.update(0, 1, 50, 2)
#     avg_acc.update(0, 2, 60, 1)
#     avg_acc.update(0, 2, 60, 2)
#
#     # task 1
#     avg_acc.update(1, 1, 90, 1)
#     avg_acc.update(1, 1, 100, 2)
#
#     # task 2
#     avg_acc.update(2, 1, 70, 1)
#     avg_acc.update(2, 1, 60, 2)
#     avg_acc.update(2, 2, 90, 1)
#     avg_acc.update(2, 2, 100, 2)
#
#     # task 3
#     avg_acc.update(3, 1, 50, 1)
#     avg_acc.update(3, 1, 50, 2)
#     avg_acc.update(3, 2, 70, 1)
#     avg_acc.update(3, 2, 60, 2)
#     avg_acc.update(3, 3, 95, 1)
#     avg_acc.update(3, 3, 100, 2)
#
#     print(avg_acc.data)
#     print(avg_acc.data.shape)
#     # print(avg_acc.compute(1))
#     print(avg_acc.compute(2))
#     # print(avg_acc.compute(3))
#     print(avg_acc.compute_final())




import torch
import numpy as np
from torch import nn
from torch import optim
from torch.nn import functional as F
from cl_gym.algorithms import ContinualAlgorithm
from cl_gym.algorithms.utils import flatten_grads, assign_grads
from cl_gym.algorithms.utils import flatten_weights, assign_weights


class MCSGD(ContinualAlgorithm):
    """
    | Mode Connectivity SGD
    | By Mirzadeh et al. :https://openreview.net/forum?id=Fmg_fQYUejf
    """
    def __init__(self, backbone, benchmark, params):
        super(MCSGD, self).__init__(backbone, benchmark, params, requires_memory=True)
        self.w_bar_prev = None
        self.w_hat_curr = None
        self.num_samples_on_line = self.params.get('mcsgd_line_samples', 5)
        self.alpha = self.params.get('mcsgd_alpha', 0.5)
    
    def calculate_line_loss(self, w_start, w_end, loader):
        line_samples = np.arange(0.0, 1.01, 1.0 / float(self.num_samples_on_line))
        accum_grad = None
        import time
        t_calc_point = 0
        t_calc_grads = 0
        for t in line_samples:
            grads = []
            w_mid = w_start + (w_end - w_start) * t
            t0 = time.time()
            m = assign_weights(self.backbone, w_mid)
            t1 = time.time()
            clf_loss = self.calculate_point_loss(m, loader)
            t2 = time.time()
            t_calc_point += (t2 - t1)
            clf_loss.backward()
            for name, param in m.named_parameters():
                grads.append(param.grad.view(-1))
            grads = torch.cat(grads)
            if accum_grad is None:
                accum_grad = grads
            else:
                accum_grad += grads
            t3 = time.time()
            t_calc_grads += (t3 - t2)
        return accum_grad/self.num_samples_on_line
   
    def calculate_point_loss(self, net, loader):
        # criterion = nn.CrossEntropyLoss()
        criterion = nn.MSELoss()
        net.eval()
        total_loss, total_count = 0.0, 0.0
        for (inp, targ, _) in loader:
            pred = net(inp)
            total_count += len(targ)
            total_loss += criterion(pred, targ)
        total_loss /= total_count
        return total_loss
    
    def find_connected_minima(self, task):
        # print(f"Debug >> w_bar_prev? {self.w_bar_prev is not None}, w_hat_curr? {self.w_hat_curr is not None}")
        mc_model = assign_weights(self.backbone, self.w_bar_prev + (self.w_hat_curr - self.w_bar_prev) * self.alpha)
        optimizer = self.prepare_optimizer(task)
        loader_prev, _ = self.benchmark.load_memory_joint(task-1, batch_size=32)
        loader_curr, _ = self.benchmark.load_subset(task, batch_size=32)
        mc_model.train()
        optimizer.zero_grad()
        grads_prev = self.calculate_line_loss(self.w_bar_prev, flatten_weights(mc_model, True), loader_prev)
        grads_curr = self.calculate_line_loss(self.w_hat_curr, flatten_weights(mc_model, True), loader_curr)
        optimizer.zero_grad()
        mc_model = assign_grads(mc_model, (grads_prev + grads_curr)/2.0)
        optimizer.step()
        return mc_model
    
    def training_epoch_end(self):
        self.w_hat_curr = flatten_weights(self.backbone, True)
        # if self.current_task > 1:
        #     self.backbone = self.find_connected_minima(self.current_task)

    def training_task_end(self):
        if self.current_task > 1:
            self.backbone = self.find_connected_minima(self.current_task)
        self.w_bar_prev = flatten_weights(self.backbone, True)
        self.current_task += 1
        
    def training_step(self, task_id, inp, targ, optimizer, criterion):
        optimizer.zero_grad()
        pred = self.backbone(inp, task_id)
        loss = criterion(pred, targ)
        loss.backward()
        # if task_id > 1:
        #     self.find_connected_minima(task_id)
        optimizer.step()



import numpy as np
from typing import Optional


class ContinualMetric:
    def __init__(self, num_tasks: int, epochs_per_task: Optional[int] = 1):
        self.num_tasks = num_tasks
        self.epochs_per_task = epochs_per_task
        # data shape => [(task_learned+1) x (task_evaluated+1) x epoch_per task]
        # The 0 index is reserved for 'initialization' metrics
        self.data = np.zeros((num_tasks+1, num_tasks+1, epochs_per_task))
    
    def update(self, task_learned: int, task_evaluated: int, value: float, epoch: Optional[int] = 1):
        if epoch < 1:
            raise ValueError("Epoch number is 1-based")
        self.data[task_learned][task_evaluated][epoch-1] = value
    
    def get_raw_history(self, task: int, start_task: Optional[int] = 1,
                        ravel: Optional[bool] = True) -> np.array:
        history = self.data[start_task:, task, :]
        if ravel:
            return np.ravel(history)
        return history
    
    def compute(self, task: int) -> float:
        raise NotImplementedError
    
    def compute_final(self) -> float:
        raise NotImplementedError




from cl_gym.algorithms.base import ContinualAlgorithm
from cl_gym.algorithms.agem import AGEM
from cl_gym.algorithms.er_ring import ERRingBuffer
from cl_gym.algorithms.ewc import EWC
from cl_gym.algorithms.ogd import OGD
from cl_gym.algorithms.mcsgd import MCSGD
from cl_gym.algorithms.mtl import Multitask

__all__ = ['ContinualAlgorithm', 'AGEM', 'ERRingBuffer', 'EWC', 'OGD', 'MCSGD', 'Multitask']



import torch
import torch.nn as nn
from typing import Iterable, Optional, Union, Iterable, Dict


class ContinualBackbone(nn.Module):
    """
    Base class for a continual backbone.
    Currently, this is a simple wrapper around PyTorch's `nn.Module` to support multiple heads.
    """
    def __init__(self,
                 multi_head: bool = False,
                 num_classes_per_head: Optional[int] = None):
        """
        Args:
            multi_head: Is this backbone multi-headed? Default: False.
            num_classes_per_head: If backbone is multi-headed, how many classes per head?
        """
        self.multi_head: bool = multi_head
        self.num_classes_per_head: int = num_classes_per_head
        if multi_head and not num_classes_per_head:
            raise ValueError("a Multi-Head Backbone is initiated without defining num_classes_per_head.")
        self.blocks: Union[Iterable[nn.Module], nn.ModuleList] = []
        super(ContinualBackbone, self).__init__()
    
    def get_block_params(self, block_id: int) -> Dict[str, torch.Tensor]:
        """
        Args:
            block_id: given the block number, provides the parameters.
        Returns:
            output: a dictionary of format {'param_name': params}
            
        . note::
            a block can have several layers (e.g., ResNet), or consist different parameters.
            For instance, the default `Linear` block has `
        """
        raise NotImplementedError
    
    @torch.no_grad()
    def get_block_outputs(self, inp: torch.Tensor, block_id: int, pre_act: bool = False) -> torch.Tensor:
        raise NotImplementedError
    
    def get_block_grads(self, block_id: int) -> torch.Tensor:
        raise NotImplementedError
     
    def select_output_head(self, output, head_ids: Iterable) -> torch.Tensor:
        """
        Helper method for selecting task-specific head.
        
        Args:
            output: The output of forward-pass. Shape: [BatchSize x ...]
            head_ids: head_ids for each example. Shape [BatchSize]

        Returns:
            output: The output where for each example in batch is calculated from one head in head_ids.
        """
        # TODO: improve performance by vectorizing this operation.
        # However, not too bad for now since number of classes is small (usually 2 or 5).
        for i, head in enumerate(head_ids):
            offset1 = int((head - 1) * self.num_classes_per_head)
            offset2 = int(head * self.num_classes_per_head)
            output[i, :offset1].data.fill_(-10e10)
            output[i, offset2:].data.fill_(-10e10)
        return output
    
    def forward(self, inp: torch.Tensor, head_ids: Optional[Iterable] = None) -> torch.Tensor:
        """
        Performs forward-pass
        
        Args:
            inp: The input tensor for forward-pass. size: [BatchSize x ...]
            head_ids: Optional list of classifier head ids. Size [BatchSize]

        Returns:
            output: Pytorch tensor of size [BatchSize x ...]
            
        . note::
            The `head_ids` will only be used if the backbone is multi-head.
        """
        out = inp
        for block in self.blocks:
            out = block(out)
        if self.multi_head:
            out = self.select_output_head(out, head_ids)
        return out



from cl_gym import benchmarks as benchmarks
from cl_gym import trainer as trainer
from cl_gym import algorithms as algorithms
from cl_gym import backbones as backbones
from cl_gym.utils import callbacks as callbacks

__version__ = "1.0.0"

__all__ = ['benchmarks',
           'trainer',
           'algorithms',
           'backbones',
           'callbacks']



from abc import ABC
from cl_gym.utils.callbacks import ContinualCallback
from typing import List


class TrainerCallbackHookMixin(ABC):

    callbacks: List[ContinualCallback] = []

    def on_before_setup(self):
        for cb in self.callbacks:
            cb.on_before_setup(self)

    def on_after_setup(self):
        for cb in self.callbacks:
            cb.on_after_setup(self)

    def on_before_teardown(self):
        for cb in self.callbacks:
            cb.on_before_teardown(self)

    def on_after_teardown(self):
        for cb in self.callbacks:
            cb.on_after_teardown(self)

    def on_before_fit(self):
        for cb in self.callbacks:
            cb.on_before_fit(self)

    def on_after_fit(self):
        for cb in self.callbacks:
            cb.on_after_fit(self)

    def on_before_training_task(self):
        for cb in self.callbacks:
            cb.on_before_training_task(self)

    def on_after_training_task(self):
        for cb in self.callbacks:
            cb.on_after_training_task(self)

    def on_before_training_epoch(self):
        for cb in self.callbacks:
            cb.on_before_training_epoch(self)

    def on_after_training_epoch(self):
        for cb in self.callbacks:
            cb.on_after_training_epoch(self)

    def on_before_training_step(self):
        for cb in self.callbacks:
            cb.on_before_training_step(self)

    def on_after_training_step(self):
        for cb in self.callbacks:
            cb.on_after_training_step(self)



from abc import ABC


class TrainerStateManagerMixin(ABC):
    """
    Handles the state (e.g., timeline, intervals) for the trainer.
    Uses book-keeping variables such as `current_step`, `current_epoch` and `current_task`.
    """
    current_step: int = 0
    current_epoch: int = 0
    current_task: int = 0
    
    def _tick_step(self):
        self.current_step += 1
    
    def _tick_epoch(self):
        self.current_epoch += 1
    
    def _tick_task(self):
        self.current_task += 1
    
    def tick(self, interval: str):
        # interval type: Literal['step', 'epoch', 'task']
        if interval == 'step':
            self._tick_step()
        elif interval == 'epoch':
            self._tick_epoch()
        elif interval == 'task':
            self._tick_task()
        else:
            raise ValueError("Supported intervals are 'step', 'epoch', and 'task")



import PIL
import torch
import torchvision
import numpy as np
from torchvision.transforms.functional import rotate

MNIST_MEAN, MNIST_STD = (0.1307,), (0.3081,)

FASHION_MNIST_MEAN, FASHION_MNIST_STD = (0.2860,), (0.3530,)

CIFAR10_MEAN, CIFAR10_STD = (0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2616)

CIFAR100_MEAN, CIFAR100_STD = (0.5071, 0.4867, 0.4408), (0.2673, 0.2564, 0.2762)


class GaussianNoiseTransform:
    def __init__(self, mean=0., std=1.):
        self.std = std
        self.mean = mean
    
    def __call__(self, tensor):
        return tensor + torch.randn(tensor.size()) * self.std + self.mean
    
    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)


def corrupt_labels(dataset, num_classes: int, corruption_prob: float):
    if isinstance(dataset.targets, list):
        target_classes = np.asarray(dataset.targets)
    else:
        target_classes = dataset.targets.clone().detach().numpy()
    mask = np.random.rand(len(target_classes)) <= corruption_prob
    random_labels = np.random.choice(num_classes, mask.sum())
    target_classes[mask] = random_labels
    target_classes = [int(x) for x in target_classes]
    dataset.targets = target_classes

 
class RotationTransform:
    """
    Rotation transforms for the images in `Rotation MNIST` dataset.
    """
    def __init__(self, angle):
        self.angle = angle
    
    def __call__(self, x):
        return rotate(x, self.angle, fill=(0,), interpolation=torchvision.transforms.InterpolationMode.BILINEAR)


class PermuteTransform:
    """
    Permutation transform, useful for permuted MNIST.
    """
    def __init__(self, permute_indices):
        self.permuted_indices = permute_indices
    
    def __call__(self, x):
        shape = x.shape
        return x.view(-1)[self.permuted_indices].view(shape)


def get_default_rotation_mnist_transform(num_tasks: int, per_task_rotation: float = None):
    if not per_task_rotation:
        per_task_rotation = 180.0/num_tasks
    transforms = []
    for task in range(1, num_tasks+1):
        rotation_degree = (task - 1) * per_task_rotation
        transform = torchvision.transforms.Compose([
            RotationTransform(rotation_degree),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(MNIST_MEAN, MNIST_STD)
        ])
        transforms.append(transform)
    assert len(transforms) == num_tasks
    return transforms


def get_default_permuted_mnist_transform(num_tasks: int):
    transforms = []
    for task in range(1, num_tasks+1):
        transform = [torchvision.transforms.ToTensor()]
        if task != 1:
            transform.append(PermuteTransform(torch.randperm(28 * 28)))
        transform.append(torchvision.transforms.Normalize(MNIST_MEAN, MNIST_STD))
        transforms.append(torchvision.transforms.Compose(transform))
    assert len(transforms) == num_tasks
    return transforms


def get_default_mnist_transform(num_tasks: int):
    transforms = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize(MNIST_MEAN, MNIST_STD),
    ])
    return [transforms]*num_tasks


def get_default_fashion_mnist_transform(num_tasks: int):
    transforms = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize(FASHION_MNIST_MEAN, FASHION_MNIST_STD),
    ])
    return [transforms]*num_tasks


def get_default_cifar_transform(num_tasks: int, is_cifar_100=False):
    normalize_mean = CIFAR100_MEAN if is_cifar_100 else CIFAR10_MEAN
    normalize_std = CIFAR100_STD if is_cifar_100 else CIFAR10_STD
    transforms = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize(normalize_mean, normalize_std),
    ])
    return [transforms]*num_tasks


def get_default_mnist_fashion_mnist_transform():
    return get_default_mnist_transform(1) + get_default_fashion_mnist_transform(1)



import os
import torch
from pathlib import Path
from cl_gym.utils.callbacks import ContinualCallback


class ModelCheckpoint(ContinualCallback):
    """
    Model Checkpoint callback.
    Saves the backbone (model) to disk at the end of every epoch or task (depending on the input settings).
    """
    def __init__(self, interval='task', name_prefix=None):
        """
        
        Args:
            interval: The intervals at which the checkpointing will be done. can be either 'task' or 'epoch'
            name_prefix: Optional prefix for model names.
        """
        if interval.lower() not in ['task', 'epoch']:
            raise ValueError("Checkpoint callback supports can only save after each 'task' or each 'epoch'")
        self.interval = interval
        self.name_prefix = name_prefix
        super(ModelCheckpoint, self).__init__()
    
    def __get_save_path(self, trainer):
        checkpoint_folder = os.path.join(trainer.params['output_dir'], 'checkpoints')
        Path(checkpoint_folder).mkdir(parents=True, exist_ok=True)
        model_name = self.name_prefix if self.name_prefix else ''
        if trainer.current_step <= 1:
            model_name += 'init.pt'
        else:
            model_name += f"{self.interval}_"
            if self.interval == 'task':
                model_name += f"{trainer.current_task}.pt"
            else:
                model_name += f"{trainer.current_epoch}.pt"
        
        filepath = os.path.join(checkpoint_folder, model_name)
        return filepath
    
    def __save_model(self, trainer):
        # get the model (backbone)
        model = trainer.algorithm.backbone
        model.eval()
        file_path = self.__get_save_path(trainer)
        torch.save(model.to('cpu').state_dict(), file_path)
    
    def on_before_fit(self, trainer):
        # save init params
        self.__save_model(trainer)
    
    def on_after_training_task(self, trainer):
        if self.interval == 'task':
            self.__save_model(trainer)
    
    def on_after_training_epoch(self, trainer):
        if self.interval == 'epoch':
            self.__save_model(trainer)


class BackboneFreeze(ContinualCallback):
    def __init__(self):
        super(BackboneFreeze, self).__init__()
    
    def _freeze_module(self, module):
        for param in module.parameters():
            param.requires_grad = False
    
    def on_after_training_task(self, trainer):
        if trainer.current_task >= 1:
            backbone = trainer.algorithm.backbone
            self._freeze_module(backbone.block_1)
            # self._freeze_module(backbone.block_2)



import torch
from typing import Optional
from sklearn.datasets import make_blobs
from torch.utils.data import DataLoader, Dataset, Subset, ConcatDataset
from cl_gym.benchmarks import Benchmark

DEFAULT_TOY_DATASET_SIZE = 100
DEFAULT_CLUSTER_STD = 0.3
SMALL_COORD = 0.5
LARGE_COORD = 2.0


class Toy2DCLFDataset(Dataset):
    def __init__(self, num_tasks: int, task_id: int, samples_per_task: int, cluster_std: float):
        self.data = []
        self.targets = []
        self.num_tasks = num_tasks
        self.task_id = task_id
        self.cluster_std = cluster_std
        self.samples_per_task = samples_per_task
        self.__generate_data()
    
    def __get_centers(self):
        if self.num_tasks == 2:
            task_centers = {1: ((SMALL_COORD, SMALL_COORD), (LARGE_COORD, LARGE_COORD)),
                            2: ((-SMALL_COORD, -SMALL_COORD), (-LARGE_COORD, -LARGE_COORD))}
        elif self.num_tasks == 4:
            task_centers = {1: ((SMALL_COORD, SMALL_COORD), (LARGE_COORD, LARGE_COORD)),
                            2: ((SMALL_COORD, -SMALL_COORD), (LARGE_COORD, -LARGE_COORD)),
                            3: ((-SMALL_COORD, -SMALL_COORD), (-LARGE_COORD, -LARGE_COORD)),
                            4: ((-SMALL_COORD, SMALL_COORD), (-LARGE_COORD, LARGE_COORD))}
        else:
            raise ValueError("2D Toy Classification dataset can have either 2 or 4 tasks")
        return task_centers
        
    def __generate_data(self):
        centers = self.__get_centers()[self.task_id]
        inp, targ = make_blobs(n_samples=self.samples_per_task, n_features=2, centers=centers, cluster_std=self.cluster_std)
        self.data = torch.from_numpy(inp).float()
        self.targets = torch.from_numpy(targ)
        
    def __getitem__(self, index):
        return self.data[index], int(self.targets[index]), self.task_id
    
    def __len__(self):
        return len(self.data)


class Toy2DClassification(Benchmark):
    """
    Toy benchmark: each task will be a binary classification with linearly separable classes in 2D space.
    Essentially, each task is a Gaussian cluster at some coordinates.
    """
    def __init__(self,
                 num_tasks: int,
                 per_task_examples: Optional[int] = None,
                 per_task_joint_examples: Optional[int] = 0,
                 per_task_memory_examples: Optional[int] = 0,
                 per_task_subset_examples: Optional[int] = 0,
                 cluster_size: Optional[int] = DEFAULT_TOY_DATASET_SIZE,
                 cluster_std: Optional[float] = DEFAULT_CLUSTER_STD):
        super().__init__(num_tasks, per_task_examples, per_task_joint_examples, per_task_memory_examples,
                         per_task_subset_examples)
        self.cluster_size = cluster_size
        self.cluster_std = cluster_std
        self.load_datasets()
        self.prepare_datasets()

    def load_datasets(self):
        for task in range(1, self.num_tasks + 1):
            self.trains[task] = Toy2DCLFDataset(self.num_tasks, task, self.cluster_size, self.cluster_std)
            self.tests[task] = Toy2DCLFDataset(self.num_tasks, task, self.cluster_size, self.cluster_std)

    def precompute_memory_indices(self):
        for task in range(1, self.num_tasks + 1):
            indices_train = self.sample_uniform_class_indices(self.trains[task], 0, 1, self.per_task_memory_examples)
            indices_test = self.sample_uniform_class_indices(self.tests[task], 0, 1, self.per_task_memory_examples)
            assert len(indices_train) == len(indices_test) == self.per_task_memory_examples
            self.memory_indices_train[task] = indices_train[:]
            self.memory_indices_test[task] = indices_test[:]



import torch
import time
from torch import nn
from torch import optim
from torch.nn import functional as F
from cl_gym.algorithms import ContinualAlgorithm


class ERRingBuffer(ContinualAlgorithm):
    """
    | Experience Replay Ring Buffer
    | By Chaudhry et al. : https://arxiv.org/abs/1902.10486.pdf
    """
    def __init__(self, backbone, benchmark, params):
        super(ERRingBuffer, self).__init__(backbone, benchmark, params, requires_memory=True)
        self.episodic_memory_iter = None
        self.episodic_memory_loader = None

    def training_step(self, task_id, inp, targ, optimizer, criterion):
        optimizer.zero_grad()
        if task_id[0] > 1:
            mem_inp, mem_targ, mem_task_ids = self.sample_batch_from_memory()
            cat_inp = torch.cat([inp, mem_inp], dim=0)
            cat_task_ids = torch.cat([task_id, mem_task_ids], dim=0)
            assert len(cat_inp) == len(cat_task_ids)
            # print(targ.shape, mem_targ.shape)
            cat_targ = torch.cat([targ, mem_targ.reshape(len(mem_targ))], dim=0)
            pred = self.backbone(cat_inp, cat_task_ids)
            loss = criterion(pred, cat_targ)
        else:
            pred = self.backbone(inp, task_id)
            loss = criterion(pred, targ)
        loss.backward()
        optimizer.step()



import torch
import torch.nn as nn
from torch.nn.functional import relu, avg_pool2d
from cl_gym.backbones import ContinualBackbone
from typing import Optional, Iterable

BN_MOMENTUM = 0.05
BN_AFFINE = True


def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, config={}):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(in_planes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes, affine=BN_AFFINE, track_running_stats=False, momentum=BN_MOMENTUM)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes, affine=BN_AFFINE, track_running_stats=False, momentum=BN_MOMENTUM)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes, affine=False, track_running_stats=False, momentum=BN_MOMENTUM)
            )

    def forward(self, x):
        out = relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = relu(out)
        return out


class ResNet(ContinualBackbone):
    def __init__(self, multi_head: bool, num_classes_per_head: int,
                 block, num_blocks, num_classes, nf, config: dict = {}):
        
        super(ResNet, self).__init__(multi_head, num_classes_per_head)
        self.in_planes = nf

        self.conv1 = conv3x3(3, nf * 1)
        self.bn1 = nn.BatchNorm2d(nf * 1, affine=BN_AFFINE, track_running_stats=False, momentum=BN_MOMENTUM)
        self.layer1 = self._make_layer(block, nf * 1, num_blocks[0], stride=1, config=config)
        self.layer2 = self._make_layer(block, nf * 2, num_blocks[1], stride=2, config=config)
        self.layer3 = self._make_layer(block, nf * 4, num_blocks[2], stride=2, config=config)
        self.layer4 = self._make_layer(block, nf * 8, num_blocks[3], stride=2, config=config)
        self.linear = nn.Linear(nf * 8 * block.expansion, num_classes)

    def _make_layer(self, block, planes, num_blocks, stride, config):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride, config=config))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, inp: torch.Tensor, head_ids: Optional[Iterable] = None):
        bsz = inp.size(0)
        out = relu(self.bn1(self.conv1(inp.view(bsz, 3, 32, 32))))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        
        if self.multi_head:
            out = self.select_output_head(out, head_ids)
        return out


class ResNet18Small(ResNet):
    """
    ResNet-18 with 1/3 less feature maps, a common backbone in CL literature.
    See Appendix C in https://openreview.net/pdf?id=Fmg_fQYUejf for further details.
    """
    def __init__(self, multi_head=True, num_classes_per_head=5, num_classes=100):
        """
        Args:
            multi_head: Is this a multi-head backbone? Default: True
            num_classes_per_head: Number of classes for each head. Default: 5 (for SplitCIFAR100)
            num_classes: total number of classes for benchmark. Default: 100 (for SplitCIFAR100)
            
        . Note::
            Since this benchmark is mostly used for Split-CIFAR100, the default arguments
            are chosen to be suitable for this benchmark.
        """
        super().__init__(multi_head, num_classes_per_head, BasicBlock, [2, 2, 2, 2], num_classes, 20)



import torch
import time
from torch import nn
from torch import optim
from cl_gym.algorithms import ContinualAlgorithm
from cl_gym.algorithms.utils import flatten_grads, assign_grads


class OGD(ContinualAlgorithm):
    """
    | Orthogonal Gradient Descent
    | By Farajtabar et al. : https://arxiv.org/abs/1910.07104.pdf
    """
    # implementation is partially based on: https://github.com/MehdiAbbanaBennani/continual-learning-ogdplus/
    def __init__(self, backbone, benchmark, params):
        super(OGD, self).__init__(backbone, benchmark, params)
        self.gradient_storage = []
        self.orthonormal_basis = None
    
    @torch.no_grad()
    def __update_orthonormal_basis(self):
        q, r = torch.qr(torch.stack(self.gradient_storage).T)
        self.orthonormal_basis = q.T
    
    @torch.no_grad()
    def __project_grad_vector(self, g):
        # print("Projection || Shape check >> g={}, basis={}".format(g.shape, self.orthonormal_basis.shape))
        mid = (torch.mm(self.orthonormal_basis, g.view(-1, 1))).T
        res = torch.mm(mid, self.orthonormal_basis)
        # print("New grad shape >> ", res.shape)
        return res.view(-1)
    
    def __update_gradient_storage(self, task):
        mem_loader_train, _ = self.benchmark.load_memory(task, batch_size=1)
        criteriton = self.prepare_criterion(self.current_task)
        optimizer = self.prepare_optimizer(self.current_task)
        self.backbone.train()
        for inp, targ, task_id in mem_loader_train:
            optimizer.zero_grad()
            pred = self.backbone(inp)
            loss = criteriton(pred, targ)
            loss.backward()
            grad_batch = flatten_grads(self.backbone).detach().clone()
            self.gradient_storage.append(grad_batch)
        optimizer.zero_grad()

    def training_task_end(self):
        self.__update_gradient_storage(self.current_task)
        self.__update_orthonormal_basis()
        
    def training_step(self, task_id, inp, targ, optimizer, criterion):
        optimizer.zero_grad()
        pred = self.backbone(inp, task_id)
        loss = criterion(pred, targ)
        loss.backward()
        pred = self.backbone(inp)
        loss = criterion(pred, targ)
        loss.backward()
        if task_id > 1:
            grad_batch = flatten_grads(self.backbone).detach().clone()
            optimizer.zero_grad()
            proj_grad = self.__project_grad_vector(grad_batch)
            new_grad = grad_batch - proj_grad
            self.backbone = assign_grads(self.backbone, new_grad)
        optimizer.step()



import os
from typing import List
from setuptools import setup, find_packages

_PATH_ROOT = os.path.dirname(__file__)


def _load_requirements(path_dir: str, file_name: str = 'requirements.txt', comment_char: str = '#') -> List[str]:
    # Load requirements from a file
    # source: https://github.com/PyTorchLightning/pytorch-lightning/blob/master/pytorch_lightning/setup_tools.py
    with open(os.path.join(path_dir, file_name), 'r') as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        # filer all comments
        if comment_char in ln:
            ln = ln[:ln.index(comment_char)].strip()
        # skip directly installed dependencies
        if ln.startswith('http') or '@http' in ln:
            continue
        if ln:  # if requirement is not empty
            reqs.append(ln)
    return reqs

setup(
  name='cl-gym',
  # packages=['cl_gym'],
  packages=find_packages(exclude=('tests', 'tests*')),
  version='1.0.3',
  include_package_data=True,
  license='MIT',
  description='Continual Learning Gym',
  author='Iman Mirzadeh',
  author_email='seyediman.mirzadeh@wsu.edu',
  url='https://github.com/imirzadeh/Cl-Gym',
  keywords=['continual learning', 'lifelong learning', 'pytorch', 'deep learning'],
  install_requires=_load_requirements(_PATH_ROOT),
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
  ],
)



import numpy as np
from typing import Optional
from cl_gym.utils.metrics import ContinualMetric


class PerformanceMetric(ContinualMetric):
    def __init__(self, num_tasks: int, epochs_per_task: Optional[int] = 1):
        super().__init__(num_tasks, epochs_per_task)
    
    def compute(self, current_task: int) -> float:
        if current_task < 1:
            raise ValueError("Tasks are 1-based. i.e., the first task's id is 1, not 0.")
        return float(np.mean(self.data[current_task, 1:current_task+1, -1]))
        
    def compute_final(self) -> float:
        return self.compute(self.num_tasks)


# if __name__ == "__main__":
#     avg_acc = AverageMetric(2, 2)
#     # init
#     avg_acc.update(0, 1, 50, 1)
#     avg_acc.update(0, 1, 50, 2)
#     avg_acc.update(0, 2, 60, 1)
#     avg_acc.update(0, 2, 60, 2)
#
#     # task 1
#     avg_acc.update(1, 1, 90, 1)
#     avg_acc.update(1, 1, 100, 2)
#
#     # task 2
#     avg_acc.update(2, 1, 70, 1)
#     avg_acc.update(2, 1, 60, 2)
#     avg_acc.update(2, 2, 90, 1)
#     avg_acc.update(2, 2, 100, 2)
#
#     print(avg_acc.data)
#     print(avg_acc.data.shape)
#     print(avg_acc.compute(1))
#     print(avg_acc.compute(2))
#     print(avg_acc.compute_final())
#     print('t1')
#     print(avg_acc.get_raw_history(1))
#     print(avg_acc.get_raw_history(1, start_task=0))
#     print('t2')
#     print(avg_acc.get_raw_history(2, start_task=0))
#     print(avg_acc.get_raw_history(2, start_task=1))
#     print(avg_acc.get_raw_history(2, start_task=2))



import torch
from typing import Dict, Iterable, Optional
from cl_gym.algorithms import ContinualAlgorithm
from cl_gym.utils.callbacks import ContinualCallback
from cl_gym.utils.loggers import Logger
from cl_gym.trainer.state_manager import TrainerStateManagerMixin
from cl_gym.trainer.callback_hooks import TrainerCallbackHookMixin


class ContinualTrainer(TrainerStateManagerMixin,
                       TrainerCallbackHookMixin):
    """
    Base class for Trainer component. Basically, orchestrates the training by implementing a state-machine.
    For further info, please see the guides on how the trainer works.
    """
    def __init__(self,
                 algorithm: ContinualAlgorithm,
                 params: dict,
                 callbacks=Iterable[ContinualCallback],
                 logger: Optional[Logger] = None):
        self.params = params
        self.algorithm = algorithm
        self.callbacks = callbacks
        self.logger = logger
    
    def setup(self):
        for cb in self.callbacks:
            cb.connect(self)
    
    def teardown(self):
        pass
    
    def train_algorithm_on_task(self, task: int):
        train_loader = self.algorithm.prepare_train_loader(task)
        optimizer = self.algorithm.prepare_optimizer(task)
        criterion = self.algorithm.prepare_criterion(task)
        device = self.params['device']
        for epoch in range(1, self.params['epochs_per_task']+1):
            self.on_before_training_epoch()
            self.tick('epoch')
            self.algorithm.backbone.train()
            self.algorithm.backbone = self.algorithm.backbone.to(device)
            for batch_idx, (inp, targ, task_ids) in enumerate(train_loader):
                self.on_before_training_step()
                self.tick('step')
                self.algorithm.training_step(task_ids.to(device), inp.to(device), targ.to(device), optimizer, criterion)
                self.algorithm.training_step_end()
                self.on_after_training_step()
            self.algorithm.training_epoch_end()
            self.on_after_training_epoch()
        self.algorithm.training_task_end()
    
    def validate_algorithm_on_task(self, task: int, validate_on_train: bool = False) -> Dict[str, float]:
        self.algorithm.backbone.eval()
        device = self.params['device']
        self.algorithm.backbone = self.algorithm.backbone.to(device)
        test_loss = 0
        correct = 0
        total = 0
        if validate_on_train:
            eval_loader = self.algorithm.prepare_train_loader(task)
        else:
            eval_loader = self.algorithm.prepare_validation_loader(task)
        criterion = self.algorithm.prepare_criterion(task)
        with torch.no_grad():
            for (inp, targ, task_ids) in eval_loader:
                inp, targ, task_ids = inp.to(device), targ.to(device), task_ids.to(device)
                pred = self.algorithm.backbone(inp, task_ids)
                total += len(targ)
                test_loss += criterion(pred, targ).item()
                pred = pred.data.max(1, keepdim=True)[1]
                correct += pred.eq(targ.data.view_as(pred)).sum()
            test_loss /= total
            correct = correct.cpu()
            avg_acc = 100.0 * float(correct.numpy()) / total
            return {'accuracy': avg_acc, 'loss': test_loss}

    def fit(self):
        for task in range(1, self.params['num_tasks']+1):
            self.on_before_training_task()
            self.tick('task')
            self.train_algorithm_on_task(task)
            self.on_after_training_task()
    
    def _run_setup(self):
        self.on_before_setup()
        self.setup()
        self.on_after_setup()
    
    def _run_fit(self):
        self.on_before_fit()
        self.fit()
        self.on_after_fit()
    
    def _run_teardown(self):
        self.on_before_teardown()
        self.teardown()
        self.on_after_teardown()

    def run(self):
        # setup
        self._run_setup()
        # fit: main training loop
        self._run_fit()
        # teardown
        self._run_teardown()



import torch
# from torch.utils.data import DataLoader, TensorDataset


class ContinualAlgorithm:
    """
    | Base class for continual learning algorithms.
    | It contains abstractions for implementing different algorithms, and also implementations shared among all algorithms.
    | It can be used for Naive(Finetune) algorithm or Stable SGD algorithm by Mirzadeh et. al.
    """
    def __init__(self, backbone, benchmark, params, requires_memory=False):
        """
        :param backbone: Neural network.
        :param benchmark: Benchmark object for preparing task training/memory data.
        :param params:
        :param requires_memory: Whether the algorithm needs episodic memory/replay buffer (e.g., A-GEM) or not.
        """
        self.backbone = backbone
        self.benchmark = benchmark
        self.params = params
        self.current_task = 1
        self.requires_memory = requires_memory
        self.episodic_memory_loader = None
        self.episodic_memory_iter = None
    
    def setup(self):
        pass
    
    def teardown(self):
        pass
    
    def update_episodic_memory(self):
        self.episodic_memory_loader, _ = self.benchmark.load_memory_joint(self.current_task,
                                                                          batch_size=self.params['batch_size_memory'],
                                                                          shuffle=True,
                                                                          pin_memory=True)
        self.episodic_memory_iter = iter(self.episodic_memory_loader)
    
    def sample_batch_from_memory(self):
        try:
            batch = next(self.episodic_memory_iter)
        except StopIteration:
            self.episodic_memory_iter = iter(self.episodic_memory_loader)
            batch = next(self.episodic_memory_iter)
        
        device = self.params['device']
        inp, targ, task_id = batch
        return inp.to(device), targ.to(device), task_id.to(device)

    def training_task_end(self):
        if self.requires_memory:
            self.update_episodic_memory()
        self.current_task += 1
    
    def training_epoch_end(self):
        pass
    
    def training_step_end(self):
        pass
    
    def prepare_train_loader(self, task_id):
        num_workers = self.params.get('num_dataloader_workers', 0)
        return self.benchmark.load(task_id, self.params['batch_size_train'],
                                   num_workers=num_workers, pin_memory=True)[0]
    
    def prepare_validation_loader(self, task_id):
        num_workers = self.params.get('num_dataloader_workers', 0)
        return self.benchmark.load(task_id, self.params['batch_size_validation'],
                                   num_workers=num_workers, pin_memory=True)[1]
    
    def prepare_optimizer(self, task_id):
        if self.params.get('learning_rate_decay'):
            lr_lower_bound = self.params.get('learning_rate_lower_bound', 0.0)
            lr = max(self.params['learning_rate'] * (self.params['learning_rate_decay'] ** (task_id-1)), lr_lower_bound)
        else:
            lr = self.params['learning_rate']
        if self.params['optimizer'].lower() == 'sgd':
            return torch.optim.SGD(self.backbone.parameters(), lr=lr, momentum=self.params['momentum'])
        elif self.params['optimizer'].lower() == 'adam':
            return torch.optim.Adam(self.backbone.parameters(), lr=lr)
        else:
            raise ValueError("Only 'SGD' and 'Adam' are accepted. To use another optimizer, override this method.")

    def prepare_criterion(self, task_id):
        return self.params['criterion']

    def training_step(self, task_ids, inp, targ, optimizer, criterion):
        optimizer.zero_grad()
        pred = self.backbone(inp, task_ids)
        loss = criterion(pred, targ)
        loss.backward()
        optimizer.step()



import os
from pathlib import Path
from typing import List, Optional


class ContinualCallback:
    """
    Base class for continual callback.
    Simply, a callback can listen to one or several events that the Trainer emits during the learning experience:
    
    .. image:: https://user-images.githubusercontent.com/8312051/122718892-4f888480-d222-11eb-87be-301e462de551.png
        :width: 400
    Please see the introduction video on what the Trainer's timeline looks like.
    
    """
    def __init__(self, name: str = '', save_dirs: Optional[List[str]] = None):
        self.name = name
        self.params = {}
        self.save_dirs = save_dirs
        self.save_paths = {}
    
    def _mkdir_save_path(self, directory_name: str):
        path = os.path.join(self.params['output_dir'], directory_name)
        Path(path).mkdir(parents=True, exist_ok=True)
        return path
    
    def connect(self, trainer):
        self.params = trainer.params
        self._prepare()
    
    def _prepare(self):
        if self.save_dirs:
            self._prepare_save_paths()
    
    def _prepare_save_paths(self):
        for directory in self.save_dirs:
            self.save_paths[directory] = self._mkdir_save_path(directory)
    
    def on_before_setup(self, trainer):
        pass

    def on_after_setup(self, trainer):
        pass

    def on_before_teardown(self, trainer):
        pass

    def on_after_teardown(self, trainer):
        pass

    def on_before_fit(self, trainer):
        pass
    
    def on_after_fit(self, trainer):
        pass
    
    def on_before_training_task(self, trainer):
        pass
    
    def on_after_training_task(self, trainer):
        pass
    
    def on_before_training_epoch(self, trainer):
        pass
    
    def on_after_training_epoch(self, trainer):
        pass
    
    def on_before_training_step(self, trainer):
        pass
    
    def on_after_training_step(self, trainer):
        pass
    
    def __repr__(self):
        return f"Callback[{self.name}]"




import torch
from torch import nn
from torch import optim
from torch.nn import functional as F
from cl_gym.algorithms import ContinualAlgorithm
from cl_gym.algorithms.utils import flatten_grads, assign_grads


class AGEM(ContinualAlgorithm):
    """
    | Averaged Gradient Episodic Memory
    | By Chaudhry et al.: https://arxiv.org/abs/1812.00420.pdf
    """
    # Implementation is partially based on: https://github.com/MehdiAbbanaBennani/continual-learning-ogdplus
    def __init__(self, backbone, benchmark, params):
        super(AGEM, self).__init__(backbone, benchmark, params, requires_memory=True)
    
    @staticmethod
    def __is_violating_direction_constraint(grad_ref, grad_batch):
        """
        GEM and A-GEM operate on gradient directions.
        i.e., gradient direction should have angle less than 90 degrees with reference gradient.
        :param grad_ref: reference gradient (i.e., grads on episodic memory)
        :param grad_batch: batch gradient
        :return:
        """
        return torch.dot(grad_ref, grad_batch) < 0
    
    @staticmethod
    def __project_grad_vector(grad_ref, grad_batch):
        """
        A-GEM operates on regularized average gradient directions.
        In case of violation, gradients should be projected (see Eq.(11) in A-GEM paper).
        :param grad_ref: reference gradient (i.e., grads on episodic memory examples)
        :param grad_batch: current batch gradients
        :return: projected gradients
        """
        return grad_batch - (torch.dot(grad_batch, grad_ref) / torch.dot(grad_ref, grad_ref)) * grad_ref
    
    def training_step(self, task_ids, inp, targ, optimizer, criterion):
        optimizer.zero_grad()
        pred = self.backbone(inp, task_ids)
        loss = criterion(pred, targ)
        loss.backward()
        if task_ids[0] > 1:
            grad_batch = flatten_grads(self.backbone).detach().clone()
            inp_ref, targ_ref, task_ids_ref = self.sample_batch_from_memory()
            loss = criterion(self.backbone(inp_ref, task_ids_ref), targ_ref.reshape(len(targ_ref)))
            loss.backward()
            grad_ref = flatten_grads(self.backbone).detach().clone()
            if self.__is_violating_direction_constraint(grad_ref, grad_batch):
                grad_ref = self.__project_grad_vector(grad_ref, grad_batch)
            optimizer.zero_grad()
            self.backbone = assign_grads(self.backbone, grad_ref)
        optimizer.step()



from cl_gym.benchmarks.base import Benchmark, DynamicTransformDataset, SplitDataset
from cl_gym.benchmarks.mnist import RotatedMNIST, PermutedMNIST, SplitMNIST
from cl_gym.benchmarks.toy_2D_clf import Toy2DClassification
from cl_gym.benchmarks.toy_1D_reg import Toy1DRegression
from cl_gym.benchmarks.pamap2 import PAMAP2
from cl_gym.benchmarks.mixed import MNISTFashionMNIST
from cl_gym.benchmarks.cifar import SplitCIFAR10, SplitCIFAR100


__all__ = ['Benchmark',
           'SplitMNIST',
           'RotatedMNIST',
           'PermutedMNIST',
           'Toy1DRegression',
           'Toy2DClassification',
           'PAMAP2',
           'MNISTFashionMNIST',
           'SplitCIFAR10',
           'SplitCIFAR100'
           ]
